"""Tests for envelope.adapters.mujoco_playground_envelope module."""

# ruff: noqa: E402

import jax
import jax.numpy as jnp
import pytest

pytestmark = pytest.mark.adapters

pytest.importorskip("mujoco_playground")

from envelope.adapters.mujoco_playground_envelope import MujocoPlaygroundEnvelope
from envelope.environment import Info
from envelope.spaces import Continuous, PyTreeSpace
from tests.contract import (
    assert_jitted_rollout_contract,
    assert_obs_matches_space,
    assert_reset_step_contract,
)


def _create_mujoco_playground_env(env_name: str = "CartpoleBalance", **kwargs):
    """Helper to create a MujocoPlaygroundEnvelope wrapper."""
    return MujocoPlaygroundEnvelope.from_name(env_name, env_kwargs=kwargs or None)


@pytest.fixture(scope="module")
def mujoco_playground_env():
    return _create_mujoco_playground_env()


@pytest.fixture(scope="module", autouse=True)
def _mujoco_playground_env_warmup(mujoco_playground_env, prng_key):
    """Warm up reset/step once to amortize compilation."""
    env = mujoco_playground_env
    key_reset, key_step = jax.random.split(prng_key)
    state, _info = env.init(key_reset)
    action = env.action_space.sample(key_step)
    env.step(state, action)


def test_mujoco_playground_contract_smoke(prng_key, mujoco_playground_env):
    assert_reset_step_contract(
        mujoco_playground_env, key=prng_key, obs_check=assert_obs_matches_space
    )


def test_mujoco_playground_contract_scan(
    prng_key, mujoco_playground_env, scan_num_steps
):
    assert_jitted_rollout_contract(
        mujoco_playground_env, key=prng_key, num_steps=scan_num_steps
    )


def test_mujoco_playground_terminated_matches_done_on_step(
    mujoco_playground_env, prng_key
):
    """MuJoCo Playground-specific: wrapper exposes underlying `done` and maps it to `terminated`."""
    env = mujoco_playground_env
    key_reset, key_action = jax.random.split(prng_key)
    state, _info = env.init(key_reset)
    action = env.action_space.sample(key_action)
    _next_state, info = env.step(state, action)
    assert hasattr(info, "done")
    assert info.terminated == info.done
    assert not info.truncated  # mujoco_playground doesn't distinguish truncation


def test_action_space_property(mujoco_playground_env):
    """Test that action_space is accessible and has correct properties."""
    env = mujoco_playground_env

    # Check that action space is a Continuous space
    assert isinstance(env.action_space, Continuous)
    assert env.action_space.shape == (env.mujoco_playground_env.action_size,)
    # MuJoCo Playground actions are typically bounded [-1, 1]
    assert jnp.allclose(env.action_space.low, -1.0)
    assert jnp.allclose(env.action_space.high, 1.0)


def test_observation_space_int_obs_size(prng_key):
    """Test observation_space for envs with int observation_size (e.g. CartpoleBalance)."""
    env = MujocoPlaygroundEnvelope.from_name("CartpoleBalance")

    # CartpoleBalance has int observation_size, should return Continuous space
    obs_size = env.mujoco_playground_env.observation_size
    assert isinstance(obs_size, int)
    assert isinstance(env.observation_space, Continuous)
    assert env.observation_space.shape == (obs_size,)

    # Verify reset produces obs that matches the space
    state, info = env.init(prng_key)
    assert env.observation_space.contains(info.obs)


def test_observation_space_dict_obs_size(prng_key):
    """Test observation_space for envs with dict observation_size (e.g. Go1Handstand)."""
    env = MujocoPlaygroundEnvelope.from_name("Go1Handstand")

    # Go1Handstand has dict observation_size, should return PyTreeSpace
    obs_size = env.mujoco_playground_env.observation_size
    assert isinstance(obs_size, dict)
    assert isinstance(env.observation_space, PyTreeSpace)

    # Check that the PyTreeSpace has the correct structure
    for key, shape in obs_size.items():
        assert key in env.observation_space.tree
        assert isinstance(env.observation_space.tree[key], Continuous)
        assert env.observation_space.tree[key].shape == shape

    # Verify reset produces obs that matches the space
    state, info = env.init(prng_key)
    assert isinstance(info.obs, dict)
    assert env.observation_space.contains(info.obs)


def test_from_name_with_env_kwargs(prng_key):
    """Test that from_name accepts env_kwargs (even if not used by registry.load)."""
    # Note: mujoco_playground.registry.load may not accept kwargs,
    # but from_name should still accept env_kwargs for consistency
    env = MujocoPlaygroundEnvelope.from_name("CartpoleBalance", env_kwargs={})

    assert env is not None
    key = prng_key
    state, info = env.init(key)
    assert state is not None


def test_episode_length_defaults_to_inf(mujoco_playground_env):
    """Test that by default episode_length is set to a very large value (effectively inf)."""
    env = mujoco_playground_env
    # Check that episode_length is set to a very large value in the config
    # (mujoco_playground uses int, so we use max int instead of inf)
    config = getattr(env.mujoco_playground_env, "_config", None)
    if config is not None:
        episode_length = config.get("episode_length")
        # Should be set to max int32 value (effectively infinite for practical purposes)
        max_int = int(jnp.iinfo(jnp.int32).max)
        assert episode_length == max_int or jnp.isposinf(jnp.asarray(episode_length))


def test_multiple_mujoco_playground_envs(prng_key):
    """Smoke-test a few different mujoco_playground environments."""
    from mujoco_playground import registry

    # Get a few different environments to test
    env_names = ["CartpoleBalance", "AcrobotSwingup", "BallInCup"]

    # Keep this deterministic and small (compile/runtime)
    for env_name in env_names:
        # Skip if environment doesn't exist
        if env_name not in registry.ALL_ENVS:
            pytest.skip(f"Environment {env_name} not available")

        env = _create_mujoco_playground_env(env_name)
        reset_key, action_key = jax.random.split(prng_key, 2)

        state, info = env.init(reset_key)
        assert state is not None
        assert isinstance(info, Info)
        # Skip expensive contains check - shape/dtype check is sufficient
        assert info.obs.shape == env.observation_space.shape

        action = env.action_space.sample(action_key)
        next_state, next_info = env.step(state, action)
        assert next_state is not None
        assert isinstance(next_info, Info)
        assert next_info.obs.shape == env.observation_space.shape
        assert jnp.all(jnp.isfinite(jnp.asarray(next_info.reward)))


def test_from_name_rejects_unknown_env_kwargs():
    """Test that from_name handles unknown/invalid env_kwargs appropriately."""
    # Try with an unknown config key - mujoco_playground may raise KeyError
    # or silently ignore it depending on implementation
    try:
        env = MujocoPlaygroundEnvelope.from_name(
            "CartpoleBalance", env_kwargs={"unknown_key": 123}
        )
        # If it doesn't raise, that's also acceptable behavior
        assert env is not None
    except (KeyError, TypeError) as e:
        # If it raises an error, that's also acceptable - just verify it's a reasonable error
        assert (
            "unknown" in str(e).lower()
            or "key" in str(e).lower()
            or "invalid" in str(e).lower()
        )


def test_config_overrides_are_passed_correctly(prng_key):
    """Test that config_overrides are correctly passed to registry.load."""
    # Test with a known config parameter (ctrl_dt)
    env = MujocoPlaygroundEnvelope.from_name(
        "CartpoleBalance", env_kwargs={"ctrl_dt": 0.072}
    )

    # Verify the config was applied
    assert env.mujoco_playground_env.dt == 0.072

    # Test with another known parameter (sim_dt)
    env2 = MujocoPlaygroundEnvelope.from_name(
        "CartpoleBalance", env_kwargs={"sim_dt": 0.005}
    )

    # Verify the config was applied
    assert env2.mujoco_playground_env.sim_dt == 0.005



"""Tests for envelope.adapters.navix_envelope module."""

# ruff: noqa: E402

import jax
import jax.numpy as jnp
import pytest

pytestmark = pytest.mark.adapters

pytest.importorskip("navix")

import navix

from envelope.adapters.navix_envelope import NavixEnvelope
from envelope.spaces import Continuous, Discrete
from tests.contract import (
    assert_jitted_rollout_contract,
    assert_obs_matches_space,
    assert_reset_step_contract,
)


def _create_navix_env(env_name: str = "Navix-Empty-5x5-v0", **kwargs):
    """Helper to create a NavixEnvelope wrapper."""
    navix_env = navix.make(env_name, **kwargs)
    return NavixEnvelope(navix_env=navix_env)


@pytest.fixture(scope="module")
def navix_env():
    return _create_navix_env()


@pytest.fixture(scope="module", autouse=True)
def _navix_env_warmup(navix_env, prng_key):
    env = navix_env
    key_reset, key_step = jax.random.split(prng_key)
    state, _info = env.init(key_reset)
    action = env.action_space.sample(key_step)
    env.step(state, action)


@pytest.mark.xfail(reason="navix obs space declares n=9 but env emits values >= 9")
def test_navix_contract_smoke(prng_key, navix_env):
    assert_reset_step_contract(
        navix_env, key=prng_key, obs_check=assert_obs_matches_space
    )


def test_navix_contract_scan(prng_key, navix_env, scan_num_steps):
    assert_jitted_rollout_contract(navix_env, key=prng_key, num_steps=scan_num_steps)


def test_action_space_conversion(navix_env):
    """Test conversion of navix action spaces to envelope spaces."""
    env = navix_env

    # Check that action space is converted correctly
    assert isinstance(env.action_space, Discrete)
    assert env.action_space.n == env.navix_env.action_space.n
    assert env.action_space.shape == env.navix_env.action_space.shape
    assert env.action_space.dtype == env.navix_env.action_space.dtype


@pytest.mark.xfail(reason="navix obs space declares n=9 but env emits values >= 9")
def test_observation_space_conversion(navix_env):
    """Test conversion of navix observation spaces to envelope spaces."""
    env = navix_env

    # Check that observation space is converted correctly
    assert isinstance(env.observation_space, Discrete)
    # Navix n might be scalar, envelope n is array - check if all elements equal navix n
    navix_n = env.navix_env.observation_space.n
    envelope_n = env.observation_space.n
    if jnp.ndim(navix_n) == 0:  # scalar
        assert jnp.all(envelope_n == navix_n)
    else:
        assert jnp.array_equal(envelope_n, navix_n)
    assert env.observation_space.shape == env.navix_env.observation_space.shape
    assert env.observation_space.dtype == env.navix_env.observation_space.dtype


def test_container_conversion_reset(navix_env, prng_key):
    """Test convert_navix_to_envelope_info on reset timestep."""
    env = navix_env
    key = prng_key

    # Get the raw navix timestep
    navix_timestep = env.navix_env.reset(key)
    state, info = env.init(key)

    # Verify obs, reward, terminated, truncated fields
    assert jnp.array_equal(info.obs, navix_timestep.observation)
    assert info.reward == navix_timestep.reward
    assert info.terminated == (navix_timestep.step_type == navix.StepType.TERMINATION)
    assert info.truncated == (navix_timestep.step_type == navix.StepType.TRUNCATION)

    # Verify terminated=False and truncated=False on reset
    assert not info.terminated
    assert not info.truncated

    # Check that extra timestep fields are preserved via update()
    # (if navix timestep has extra fields beyond the standard ones)
    import dataclasses

    timestep_dict = dataclasses.asdict(navix_timestep)
    # Remove standard fields that are handled explicitly
    standard_fields = {"observation", "reward", "step_type"}
    extra_fields = {k: v for k, v in timestep_dict.items() if k not in standard_fields}
    for field_name, field_value in extra_fields.items():
        assert hasattr(info, field_name)


def test_container_conversion_step(navix_env, prng_key):
    """Test convert_navix_to_envelope_info on step timestep."""
    env = navix_env
    key = prng_key

    state, _ = env.init(key)
    action = env.action_space.sample(jax.random.fold_in(prng_key, 1))

    # Get the raw navix timestep
    navix_timestep = env.navix_env.step(state, action)
    next_state, info = env.step(state, action)

    # Verify all fields are correctly converted
    assert jnp.array_equal(info.obs, navix_timestep.observation)
    assert info.reward == navix_timestep.reward
    assert info.terminated == (navix_timestep.step_type == navix.StepType.TERMINATION)
    assert info.truncated == (navix_timestep.step_type == navix.StepType.TRUNCATION)

    # Verify reward values are preserved
    assert info.reward == navix_timestep.reward


def test_episode_truncation(prng_key):
    """Test that episode truncation is correctly detected."""
    env = _create_navix_env(max_steps=5)  # Very short episode
    key = prng_key

    state, info = env.init(key)

    # Step until truncation
    for _ in range(10):  # More than max_steps
        if info.truncated:
            break
        action = env.action_space.sample(jax.random.fold_in(prng_key, _))
        state, info = env.step(state, action)

    # Verify truncation occurred
    assert info.truncated
    assert not info.terminated


def test_unsupported_space_type():
    """Test that unsupported space types raise ValueError."""
    import jax.numpy as jnp
    from navix import spaces as navix_spaces

    from envelope.adapters.navix_envelope import convert_navix_to_envelope_space

    # Create a mock space that's neither Discrete nor Continuous
    class MockSpace(navix_spaces.Space):
        pass

    # Provide required arguments for Space base class
    mock_space = MockSpace(
        shape=(),
        dtype=jnp.int32,
        minimum=jnp.array(0),
        maximum=jnp.array(10),
    )
    with pytest.raises(ValueError, match="Unsupported space type"):
        convert_navix_to_envelope_space(mock_space)


def test_step_type_conversion(navix_env, prng_key):
    """Test all navix StepType values are correctly converted."""
    import navix

    from envelope.adapters.navix_envelope import convert_navix_to_envelope_info

    env = navix_env
    key = prng_key

    # Test TRANSITION step type (should be on reset)
    reset_timestep = env.navix_env.reset(key)
    reset_info = convert_navix_to_envelope_info(reset_timestep)
    # TRANSITION should map to neither terminated nor truncated
    assert reset_timestep.step_type == navix.StepType.TRANSITION
    assert not reset_info.terminated
    assert not reset_info.truncated

    # Test TRANSITION on normal step (before termination/truncation)
    state, _ = env.init(key)
    action = env.action_space.sample(jax.random.fold_in(prng_key, 1))
    step_timestep = env.navix_env.step(state, action)
    step_info = convert_navix_to_envelope_info(step_timestep)
    # TRANSITION should map to neither terminated nor truncated
    if step_timestep.step_type == navix.StepType.TRANSITION:
        assert not step_info.terminated
        assert not step_info.truncated

    # Verify the conversion logic works correctly for all step types
    assert step_info.terminated == (
        step_timestep.step_type == navix.StepType.TERMINATION
    )
    assert step_info.truncated == (step_timestep.step_type == navix.StepType.TRUNCATION)

    # Test that TERMINATION maps correctly
    # (We can't easily trigger termination, but we verify the logic)
    assert step_info.terminated == (
        int(step_timestep.step_type) == int(navix.StepType.TERMINATION)
    )
    assert step_info.truncated == (
        int(step_timestep.step_type) == int(navix.StepType.TRUNCATION)
    )


def test_discrete_space_conversion():
    """Test conversion of discrete spaces from navix to envelope."""
    from navix import spaces as navix_spaces

    from envelope.adapters.navix_envelope import convert_navix_to_envelope_space

    # Create a navix Discrete space
    navix_discrete = navix_spaces.Discrete.create(10, shape=(3,), dtype=jnp.int32)
    envelope_discrete = convert_navix_to_envelope_space(navix_discrete)
    assert isinstance(envelope_discrete, Discrete)
    # Navix n might be scalar, envelope n can be broadcast to shape.
    navix_n = navix_discrete.n
    envelope_n = envelope_discrete.n
    if jnp.ndim(navix_n) == 0:
        assert jnp.all(envelope_n == navix_n)
    else:
        assert jnp.array_equal(envelope_n, navix_n)
    assert envelope_discrete.shape == navix_discrete.shape
    assert envelope_discrete.dtype == navix_discrete.dtype


def test_continuous_space_conversion():
    """Test conversion of continuous spaces from navix to envelope."""
    from navix import spaces as navix_spaces

    from envelope.adapters.navix_envelope import convert_navix_to_envelope_space

    # Create a navix Continuous space
    navix_continuous = navix_spaces.Continuous.create(
        shape=(3,),
        minimum=jnp.array([-1.0, -2.0, -3.0]),
        maximum=jnp.array([1.0, 2.0, 3.0]),
        dtype=jnp.float32,
    )

    # Convert to envelope space
    envelope_continuous = convert_navix_to_envelope_space(navix_continuous)

    # Verify conversion
    assert isinstance(envelope_continuous, Continuous)
    assert envelope_continuous.shape == navix_continuous.shape
    assert envelope_continuous.dtype == navix_continuous.dtype
    assert jnp.array_equal(envelope_continuous.low, navix_continuous.minimum)
    assert jnp.array_equal(envelope_continuous.high, navix_continuous.maximum)



from typing import override

import jax
import jax.numpy as jnp

from envelope.environment import Info, State
from envelope.spaces import BatchedSpace, Continuous, Discrete, PyTreeSpace, Space
from envelope.typing import PyTree
from envelope.wrappers.wrapper import Wrapper


def clip_action(action: PyTree, space: Space) -> PyTree:
    if isinstance(space, BatchedSpace):
        return jax.vmap(clip_action, in_axes=(0, None))(action, space.space)
    elif isinstance(space, PyTreeSpace):
        return jax.tree.map(clip_action, action, space.tree)
    elif isinstance(space, Continuous):
        return jnp.clip(action, space.low, space.high)
    elif isinstance(space, Discrete):
        return jnp.clip(action, 0, space.n - 1)


class ClipActionWrapper(Wrapper):
    @override
    def step(self, state: State, action: PyTree) -> tuple[State, Info]:
        action = clip_action(action, self.action_space)
        return self.env.step(state, action)



from typing import Any, Protocol, TypeAlias, runtime_checkable

import jax

PyTree: TypeAlias = Any
Key: TypeAlias = jax.Array  # with jnp.issubdtype(key.dtype, jax.dtypes.prng_key)
Array: TypeAlias = jax.Array
State: TypeAlias = PyTree


@runtime_checkable
class Info(Protocol):
    """
    `Info` is a runtime-checkable Protocol that defines required fields and methods for
    environment emissions, including observation, reward, and termination/truncation
    flags.

    Attributes:
        obs: The observation from the environment.
        reward: The reward from the environment.
        terminated: Whether the episode has terminated.
        truncated: Whether the episode has truncated.

    """

    obs: PyTree
    reward: float
    terminated: bool
    truncated: bool

    def update(self, **changes: PyTree) -> "Info":
        """Update the info container with new values. This method should return
        a new instance with updated and potentially new values."""
        ...

    def __getattr__(self, name: str) -> PyTree: ...



from abc import ABC, abstractmethod
from dataclasses import field
from functools import cached_property

from envelope import spaces
from envelope.struct import Container, FrozenPyTreeNode
from envelope.typing import Array, Info, Key, PyTree, State

__all__ = ["Environment", "InfoContainer"]


class InfoContainer(Container):
    """
    Info container for environment emissions, including observation, reward, and
    termination/truncation flags. This container implements the `Info` protocol.

    Attributes:
        obs: The observation from the environment.
        reward: The reward from the environment.
        terminated: Whether the episode has terminated.
        truncated: Whether the episode has truncated.
    """

    obs: PyTree
    reward: float | Array
    terminated: bool
    truncated: bool = field(default=False)


class Environment(ABC, FrozenPyTreeNode):
    """
    Base class for all environments.

    State is an opaque PyTree owned by each environment; wrappers that stack
    environments should expose their wrapped env state as `inner_state` while
    adding any wrapper-specific fields.
    """

    @abstractmethod
    def init(self, key: Key) -> tuple[State, Info]:
        """
        Initialize environment state and sample the initial info.

        This method closely resembles the `reset` method of gymnasium or gymnax.
        However, it is normally called only once per environment lifecycle, as
        subsequent resets should be performed using the `Environment.reset` method,
        which may preserve episode-persistent state.
        """
        ...

    def reset(self, state: State, key: Key) -> tuple[State, Info]:
        """
        Reset the environment while preserving episode-persistent state.

        By default, `state` is ignored and `init` is called, as this most closely
        matches the standard reinforcement learning setting in which the starting state
        is sampled from a fixed distribution.
        """
        return self.init(key)

    @abstractmethod
    def step(self, state: State, action: PyTree) -> tuple[State, Info]:
        """Step the environment given an action, returning the next state and info."""
        ...

    @abstractmethod
    @cached_property
    def observation_space(self) -> spaces.Space:
        """The space of observations."""
        ...

    @abstractmethod
    @cached_property
    def action_space(self) -> spaces.Space:
        """The space of actions."""
        ...

    @property
    def unwrapped(self) -> "Environment":
        return self



from envelope.adapters import create
from envelope.environment import Environment, Info, InfoContainer
from envelope.spaces import BatchedSpace, Continuous, Discrete, PyTreeSpace, Space
from envelope.struct import Container, FrozenPyTreeNode, field, static_field
from envelope.wrappers import (
    AutoResetWrapper,
    ClipActionWrapper,
    ContinuousObservationWrapper,
    EpisodeStatisticsWrapper,
    FlattenActionWrapper,
    FlattenObservationWrapper,
    ObservationNormalizationWrapper,
    PooledInitVmapWrapper,
    StateInjectionWrapper,
    TruncationWrapper,
    VmapEnvsWrapper,
    VmapWrapper,
    WrappedState,
    Wrapper,
)

__all__ = [
    # Basic functionality
    "create",
    "Environment",
    "Info",
    "InfoContainer",
    # Spaces
    "Space",
    "BatchedSpace",
    "Continuous",
    "Discrete",
    "PyTreeSpace",
    # Struct
    "field",
    "static_field",
    "FrozenPyTreeNode",
    "Container",
    # Wrappers
    "Wrapper",
    "WrappedState",
    "AutoResetWrapper",
    "ClipActionWrapper",
    "ContinuousObservationWrapper",
    "EpisodeStatisticsWrapper",
    "FlattenActionWrapper",
    "FlattenObservationWrapper",
    "ObservationNormalizationWrapper",
    "PooledInitVmapWrapper",
    "StateInjectionWrapper",
    "TruncationWrapper",
    "VmapWrapper",
    "VmapEnvsWrapper",
]



from functools import cached_property
from typing import override

import jax
import jax.numpy as jnp

from envelope import spaces
from envelope.environment import Info
from envelope.struct import field
from envelope.typing import Key, PyTree
from envelope.wrappers.vmap_wrapper import _split_or_keep_key
from envelope.wrappers.wrapper import WrappedState, Wrapper


class PooledInitVmapWrapper(Wrapper):
    batch_size: int = field(kw_only=True)
    pool_size: int = field(kw_only=True)

    class PooledInitVmapState(WrappedState):
        init_key: Key = field()
        last_final: Info = field()

    @override
    def init(self, key: Key) -> tuple[WrappedState, Info]:
        keys = _split_or_keep_key(key, self.batch_size + 1)
        key_next, keys_pool = keys[0], keys[1:]
        inner_state, info = jax.vmap(self.env.init)(keys_pool)

        # Note that the placeholder that is returned only has nan values where it's
        # dtype is a subdtype of float. TODO: Should we use empty_like?
        pholder_info = jax.tree.map(lambda x: jnp.full_like(x, jnp.nan), info)
        state = self.PooledInitVmapState(
            inner_state=inner_state,
            init_key=key_next,
            last_final=pholder_info,
        )
        return state, info.update(final=pholder_info)

    @override
    def reset(self, state: WrappedState, key: Key) -> tuple[WrappedState, Info]:
        # It's hard to support reset for this wrapper.
        # We would have to init the state of a pool of unwrapped environments, and then
        # somehow inject this into the stack of wrapped states. The current data
        # structure for wrapped states does not make this possible without being super
        # hacky, and violating the assumption that wrapped states are opaque (we would
        # likely have to recursively descend by checking if
        # hasattr(state, "inner_state")).
        # Since there is currently no use case in which we need to carry state across
        # episodes before vmapping, we will implement this later.
        keys = _split_or_keep_key(key, self.batch_size + 1)
        key_next, keys_pool = keys[0], keys[1:]
        inner_state, info = jax.vmap(self.env.reset)(state.inner_state, keys_pool)
        state = state.replace(inner_state=inner_state, init_key=key_next)
        return state, info.update(final=state.last_final)

    @override
    def step(self, state: WrappedState, action: PyTree) -> tuple[WrappedState, Info]:
        inner_state, info = jax.vmap(self.env.step)(state.inner_state, action)
        done = info.terminated | info.truncated

        # Compute pool_size fresh init states
        key_pool = jax.random.fold_in(state.init_key, 0)
        next_init_key = jax.random.fold_in(state.init_key, 1)
        keys_pool = jax.random.split(key_pool, self.pool_size)
        inner_states_pool, infos_pool = jax.vmap(self.env.init)(keys_pool)

        # Randomly assign each env a init state from the pool
        key_idxs = jax.random.fold_in(state.init_key, 2)
        pool_idxs = jax.random.randint(key_idxs, (self.batch_size,), 0, self.pool_size)

        # Expand pool states to batch_size via indexing
        mapped_init_state = jax.tree.map(lambda x: x[pool_idxs], inner_states_pool)
        mapped_init_info = jax.tree.map(lambda x: x[pool_idxs], infos_pool)

        # Select inner_state: init for done envs, continue for others
        final_inner_state = jax.tree.map(
            lambda init, curr: jax.vmap(jnp.where)(done, init, curr),
            mapped_init_state,
            inner_state,
        )

        # Select last_final: on done, store terminal info; on continue, keep previous
        final_last_final = jax.tree.map(
            lambda curr, prev: jax.vmap(jnp.where)(done, curr, prev),
            info,
            state.last_final,
        )

        # Build final_info with final field
        # For done envs: obs is new initial obs, final is terminal info
        # For continue envs: obs is current obs, final is previous last_final
        final_obs = jax.tree.map(
            lambda init, curr: jax.vmap(jnp.where)(done, init, curr),
            mapped_init_info.obs,
            info.obs,
        )
        final_final = jax.tree.map(
            lambda curr, prev: jax.vmap(jnp.where)(done, curr, prev),
            info,  # Terminal info snapshot for done envs
            state.last_final,  # Previous episode's final for continue envs
        )
        final_info = info.update(obs=final_obs, final=final_final)

        state = state.replace(
            inner_state=final_inner_state,
            init_key=next_init_key,
            last_final=final_last_final,
        )
        return state, final_info

    @override
    @cached_property
    def observation_space(self) -> spaces.Space:
        return spaces.BatchedSpace(
            space=self.env.observation_space, batch_size=self.batch_size
        )

    @override
    @cached_property
    def action_space(self) -> spaces.Space:
        return spaces.BatchedSpace(
            space=self.env.action_space, batch_size=self.batch_size
        )



import jax
import jax.numpy as jnp
import pytest

from envelope.wrappers.truncation_wrapper import TruncationWrapper
from tests.wrappers.helpers import DiscreteStepCounterEnv, StepCounterEnv


def test_reset_sets_truncated_false():
    env = StepCounterEnv()
    w = TruncationWrapper(env, max_steps=3)
    key = jax.random.key(0)
    state, info = w.init(key)
    assert state is not None
    assert info.truncated is False


@pytest.mark.parametrize(
    "env_factory,actions,max_steps,expected_truncated_seq",
    [
        (StepCounterEnv, [0.1, 0.2, 0.3, 0.4], 3, [False, False, True, True]),
        (StepCounterEnv, [0.1], 1, [True]),
        (DiscreteStepCounterEnv, [1, 2, 3], 2, [False, True, True]),
    ],
    ids=["cont_ms3", "cont_ms1", "disc_ms2"],
)
def test_step_truncates_at_threshold(
    env_factory, actions, max_steps, expected_truncated_seq
):
    env = env_factory()
    w = TruncationWrapper(env, max_steps=max_steps)
    key = jax.random.key(0)
    state, info = w.init(key)
    truncs = []
    for a in actions:
        state, info = w.step(state, jnp.asarray(a))
        truncs.append(bool(info.truncated))
    assert truncs == expected_truncated_seq


def test_preserves_other_info_fields():
    env = StepCounterEnv()
    w = TruncationWrapper(env, max_steps=2)
    key = jax.random.key(0)
    state, info = w.init(key)
    # Step once: not truncated yet
    state, info = w.step(state, jnp.asarray(0.5))
    assert info.terminated is False
    assert jnp.allclose(info.obs, state.unwrapped.env_state)
    # Step twice: hits threshold
    state, info = w.step(state, jnp.asarray(-0.25))
    assert info.terminated is False
    assert jnp.allclose(info.obs, state.unwrapped.env_state)
    assert bool(jnp.asarray(info.truncated)) is True


def test_reset_overrides_underlying_truncated_true():
    env = StepCounterEnv(reset_truncated=True)
    w = TruncationWrapper(env, max_steps=5)
    key = jax.random.key(0)
    state, info = w.init(key)
    assert info.truncated is False


@pytest.mark.parametrize("max_steps", [0, 1])
def test_max_steps_edge_values(max_steps):
    env = StepCounterEnv()
    w = TruncationWrapper(env, max_steps=max_steps)
    key = jax.random.key(0)
    state, info = w.init(key)
    # First step should truncate immediately when max_steps == 0
    state, info = w.step(state, jnp.asarray(0.0))
    # After first step, steps == 1; wrapper truncates when steps >= max_steps
    expected = 1 >= max_steps
    assert bool(jnp.asarray(info.truncated)) == expected


def test_truncated_remains_true_after_threshold():
    env = StepCounterEnv()
    w = TruncationWrapper(env, max_steps=2)
    key = jax.random.key(0)
    state, info = w.init(key)
    # Step 1: steps=1 < 2
    state, info = w.step(state, jnp.asarray(0.1))
    assert bool(jnp.asarray(info.truncated)) is False
    # Step 2: steps=2 == 2 -> truncated
    state, info = w.step(state, jnp.asarray(0.1))
    assert bool(jnp.asarray(info.truncated)) is True
    # Step 3: stays truncated
    state, info = w.step(state, jnp.asarray(0.1))
    assert bool(jnp.asarray(info.truncated)) is True


def test_wrapper_overrides_underlying_truncated_on_step():
    env = StepCounterEnv(step_truncated=True)
    w = TruncationWrapper(env, max_steps=10)
    key = jax.random.key(0)
    state, info = w.init(key)
    # Underlying env sets truncated True, but steps < max_steps => wrapper should set False
    state, info = w.step(state, jnp.asarray(0.5))
    assert bool(jnp.asarray(info.truncated)) is False


def test_steps_as_jax_scalar_array_behaves_correctly():
    env = StepCounterEnv(steps_dtype=jnp.int32)
    w = TruncationWrapper(env, max_steps=2)
    key = jax.random.key(0)
    state, info = w.init(key)
    # After one step: steps = 1 (jax scalar), not truncated
    state, info = w.step(state, jnp.asarray(0.1))
    assert jnp.asarray(info.truncated).dtype == jnp.bool_
    assert bool(jnp.asarray(info.truncated)) is False
    # After second step: steps = 2 -> truncated
    state, info = w.step(state, jnp.asarray(0.1))
    assert bool(jnp.asarray(info.truncated)) is True


def test_reset_with_state_passes_inner_state_down():
    """reset(state, key) should pass state.inner_state to the inner env's reset."""
    env = StepCounterEnv()
    w = TruncationWrapper(env, max_steps=10)
    key = jax.random.key(0)

    state, _ = w.init(key)
    for _ in range(5):
        state, _ = w.step(state, jnp.asarray(0.1))
    assert state.steps == 5

    new_state, _ = w.reset(state, jax.random.key(1))

    # Inner env should be reset
    assert jnp.allclose(new_state.inner_state.env_state, 0.0)
    # Steps is episode-scoped, so it resets to 0
    assert new_state.steps == 0


def test_init_does_not_pass_inner_state():
    """init(key) should not pass inner state to the inner env."""
    env = StepCounterEnv()
    w = TruncationWrapper(env, max_steps=10)
    key = jax.random.key(0)

    state, _ = w.init(key)
    assert state.steps == 0


@pytest.mark.parametrize(
    "env_factory,action",
    [
        (StepCounterEnv, jnp.asarray(0.3)),
        (DiscreteStepCounterEnv, jnp.asarray(3, dtype=jnp.int32)),
    ],
    ids=["jit-cont", "jit-disc"],
)
def test_jit_compatibility(env_factory, action):
    env = env_factory()
    w = TruncationWrapper(env, max_steps=2)
    key = jax.random.key(0)

    # Avoid returning InfoContainer across JIT boundary; return only needed pieces
    reset_jit_state = jax.jit(lambda k: w.init(k)[0])
    step_jit_state_trunc = jax.jit(
        lambda s, a: (w.step(s, a)[0], w.step(s, a)[1].truncated)
    )

    state = reset_jit_state(key)

    next_state, truncated = step_jit_state_trunc(state, action)
    # `truncated` may be a JAX scalar array after JIT; validate dtype/shape
    assert jnp.asarray(truncated).dtype == jnp.bool_
    assert jnp.shape(truncated) == ()



"""Tests for envelope.adapters.craftax_envelope module."""

# ruff: noqa: E402

import jax
import jax.numpy as jnp
import pytest

pytestmark = pytest.mark.adapters

pytest.importorskip("craftax")

from craftax.craftax.envs.craftax_pixels_env import (
    CraftaxPixelsEnv,
    CraftaxPixelsEnvNoAutoReset,
)
from craftax.craftax_classic.envs.craftax_pixels_env import (
    CraftaxClassicPixelsEnv,
    CraftaxClassicPixelsEnvNoAutoReset,
)

from envelope.spaces import Continuous, Discrete
from tests.contract import (
    assert_jitted_rollout_contract,
    assert_obs_matches_space,
    assert_reset_step_contract,
)


@pytest.fixture(
    params=[
        "Craftax-Symbolic-v1",
        "Craftax-Classic-Symbolic-v1",
        "Craftax-Pixels-v1",
        "Craftax-Classic-Pixels-v1",
    ],
    ids=["symbolic", "classic_symbolic", "pixels", "classic_pixels"],
    scope="module",
)
def craftax_env_id(request: pytest.FixtureRequest) -> str:
    return request.param


@pytest.fixture(scope="module")
def craftax_env(craftax_env_id: str):
    from envelope.adapters.craftax_envelope import CraftaxEnvelope

    return CraftaxEnvelope.from_name(craftax_env_id)


@pytest.fixture(scope="module", autouse=True)
def _craftax_env_warmup(craftax_env, prng_key):
    """Warm up reset/step once per Craftax variant to amortize compilation."""
    env = craftax_env
    key_reset, key_step = jax.random.split(prng_key)
    state, _info = env.init(key_reset)
    action = env.action_space.sample(key_step)
    env.step(state, action)


def _one_step(env, state, key):
    action = env.action_space.sample(key)
    return env.step(state, action)


def test_craftax_contract_smoke(craftax_env, prng_key):
    failing_envs = (
        CraftaxPixelsEnv,
        CraftaxClassicPixelsEnv,
        CraftaxPixelsEnvNoAutoReset,
        CraftaxClassicPixelsEnvNoAutoReset,
    )
    if isinstance(craftax_env.craftax_env, failing_envs):
        pytest.xfail(
            "Craftax currently returns the wrong observation space. "
            "See https://github.com/MichaelTMatthews/Craftax/pull/49"
        )
    assert_reset_step_contract(
        craftax_env, key=prng_key, obs_check=assert_obs_matches_space
    )


def test_craftax_contract_scan(craftax_env, prng_key, scan_num_steps):
    assert_jitted_rollout_contract(craftax_env, key=prng_key, num_steps=scan_num_steps)


def test_spaces_exposed(craftax_env):
    assert craftax_env.action_space is not None
    assert craftax_env.observation_space is not None
    assert isinstance(craftax_env.action_space, Discrete)
    assert isinstance(craftax_env.observation_space, Continuous)


def test_time_limit_overridden_to_inf(craftax_env):
    if hasattr(craftax_env.env_params, "max_timesteps"):
        assert jnp.isposinf(jnp.asarray(craftax_env.env_params.max_timesteps))


def test_key_splitting_reset_and_step(craftax_env, prng_key):
    key = prng_key

    state, _ = craftax_env.init(key)
    assert not jnp.array_equal(state.key, key)

    _key_step = jax.random.fold_in(prng_key, 1)
    next_state, _ = _one_step(craftax_env, state, _key_step)
    assert not jnp.array_equal(next_state.key, state.key)


class _DummyParams:
    def __init__(self, max_timesteps):
        self.max_timesteps = max_timesteps

    def replace(self, **updates):
        return _DummyParams(updates.get("max_timesteps", self.max_timesteps))


class _DummyEnv:
    def __init__(self, default_params):
        self.default_params = default_params


def test_from_name_errors_on_auto_reset():
    from envelope.adapters.craftax_envelope import CraftaxEnvelope

    with pytest.raises(ValueError, match="Cannot override 'auto_reset' directly"):
        CraftaxEnvelope.from_name("AnyEnv", env_kwargs={"auto_reset": True})



import pickle

import jax
import jax.numpy as jnp
import pytest
from hypothesis import given, settings
from hypothesis import strategies as st

from envelope.environment import Info
from envelope.spaces import BatchedSpace
from envelope.wrappers.observation_normalization_wrapper import (
    ObservationNormalizationWrapper,
)
from envelope.wrappers.vmap_wrapper import VmapWrapper
from tests.wrappers.helpers import FlagDoneEnv, ScalarToyEnv, VectorToyEnv

# -----------------------------------------------------------------------------
# Core: Space shaping and protocol conformance
# -----------------------------------------------------------------------------


@pytest.mark.parametrize("batch_size", [1, 2, 5])
def test_spaces_are_batched(batch_size):
    env = ScalarToyEnv()
    wrapped = VmapWrapper(env, batch_size=batch_size)
    assert isinstance(wrapped.observation_space, BatchedSpace)
    assert isinstance(wrapped.action_space, BatchedSpace)
    assert (
        wrapped.observation_space.shape == (batch_size,) + env.observation_space.shape
    )
    assert wrapped.action_space.shape == (batch_size,) + env.action_space.shape


def test_protocol_conformance_reset_and_step():
    env = ScalarToyEnv()
    wrapped = VmapWrapper(env, batch_size=3)
    key = jax.random.key(0)
    state, info = wrapped.init(key)
    assert isinstance(info, Info)
    assert state is not None
    # Action space contains batched action
    action = wrapped.action_space.sample(key)
    assert wrapped.action_space.contains(action)
    next_state, next_info = wrapped.step(state, action)
    assert isinstance(next_info, Info)
    assert next_state is not None
    assert wrapped.observation_space.contains(next_info.obs)


# -----------------------------------------------------------------------------
# Core: Reset/step equivalence vs manual vmap
# -----------------------------------------------------------------------------


@pytest.mark.parametrize("batch_size", [2, 4])
def test_reset_equivalence_to_manual_vmap(batch_size):
    env = ScalarToyEnv()
    w = VmapWrapper(env, batch_size=batch_size)
    single_key = jax.random.key(0)
    keys = jax.random.split(single_key, batch_size)
    s_wrapped, i_wrapped = w.init(single_key)
    s_manual, i_manual = jax.vmap(env.init)(keys)
    assert jax.tree_util.tree_all(
        jax.tree.map(lambda a, b: jnp.allclose(a, b), s_wrapped, s_manual)
    )
    assert jax.tree_util.tree_all(
        jax.tree.map(lambda a, b: jnp.allclose(a, b), i_wrapped.obs, i_manual.obs)
    )


@pytest.mark.parametrize("dim,batch_size", [(1, 3), (4, 2)])
def test_step_equivalence_to_manual_vmap(dim, batch_size):
    env = VectorToyEnv(dim)
    w = VmapWrapper(env, batch_size=batch_size)
    key = jax.random.key(0)
    s0, _ = w.init(key)
    action = w.action_space.sample(key)
    s1, info1 = w.step(s0, action)
    # Manual baseline
    keys = jax.random.split(key, batch_size)
    s0_m, _ = jax.vmap(env.init)(keys)
    s1_m, info1_m = jax.vmap(env.step)(s0_m, action)
    assert jax.tree_util.tree_all(
        jax.tree.map(lambda a, b: jnp.allclose(a, b), s1, s1_m)
    )
    assert jnp.allclose(info1.reward, info1_m.reward)
    assert jnp.allclose(info1.obs, info1_m.obs)


# -----------------------------------------------------------------------------
# Core: Error paths
# -----------------------------------------------------------------------------


def test_reset_raises_on_wrong_key_batch_dim():
    env = ScalarToyEnv()
    w = VmapWrapper(env, batch_size=4)
    bad_keys = jax.random.split(jax.random.key(0), 3)  # mismatch batch
    with pytest.raises(ValueError):
        _ = w.init(bad_keys)


def test_step_raises_on_action_shape_mismatch():
    env = VectorToyEnv(dim=3)
    w = VmapWrapper(env, batch_size=2)
    key = jax.random.key(0)
    state, _ = w.init(key)
    bad_action = jnp.ones((3,), dtype=jnp.float32)  # missing batch dim
    with pytest.raises(Exception):
        _ = w.step(state, bad_action)


# -----------------------------------------------------------------------------
# Core: Composability with normalization and order equivalence
# -----------------------------------------------------------------------------


@pytest.mark.parametrize("batch_size", [2, 5])
def test_normalize_then_vmap_equals_vmap_then_normalize(batch_size):
    base = VectorToyEnv(dim=3)
    # Order A: normalize then vmap
    a = VmapWrapper(
        ObservationNormalizationWrapper(base),
        batch_size=batch_size,
    )
    # Order B: vmap then normalize
    b = ObservationNormalizationWrapper(VmapWrapper(base, batch_size=batch_size))
    key = jax.random.key(42)
    s_a, i_a = a.init(key)
    s_b, i_b = b.init(key)
    # Both produce batched observations with same shape/dtype; numerical values
    # may differ due to per-env vs aggregated RMV semantics
    assert i_a.obs.shape == i_b.obs.shape == (batch_size, 3)
    assert i_a.obs.dtype == i_b.obs.dtype

    # Check normalization statistics shapes:
    # Order A (normalize then vmap): rmv_state is vmapped, so each env has its own stats
    rmv_a = s_a.rmv_state
    assert rmv_a.mean.shape == (batch_size, 3)
    assert rmv_a.var.shape == (batch_size, 3)

    # Order B (vmap then normalize): rmv_state is shared (unbatched)
    rmv_b = s_b.rmv_state
    assert rmv_b.mean.shape == (3,)
    assert rmv_b.var.shape == (3,)


def test_nested_vmaps_equivalence_reset_and_step():
    base = VectorToyEnv(dim=2)
    outer = VmapWrapper(VmapWrapper(base, batch_size=3), batch_size=2)  # (2,3)
    flat = VmapWrapper(base, batch_size=6)  # (6,)
    key = jax.random.key(0)
    s_outer, i_outer = outer.init(key)
    keys = jax.random.split(key, 6)
    s_flat_m, i_flat_m = jax.vmap(base.init)(keys)
    # Reshape outer to flat
    i_outer_flat = i_outer.obs.reshape((6, 2))
    assert jnp.allclose(i_outer_flat, i_flat_m.obs)
    action = flat.action_space.sample(key)
    s_flat, _ = flat.init(key)
    s_flat2, i_flat2 = flat.step(s_flat, action)
    s_outer2, i_outer2 = outer.step(s_outer, action.reshape((2, 3, 2)))
    i_outer2_flat = i_outer2.obs.reshape((6, 2))
    assert jnp.allclose(i_outer2_flat, i_flat2.obs)


def test_termination_truncation_propagation():
    flags = jnp.array([True, False, True], dtype=bool)
    w = VmapWrapper(FlagDoneEnv(flags=flags), batch_size=3)
    s, _ = w.init(jax.random.key(0))
    _, info = w.step(s, jnp.array([0.1, 0.2, 0.3], dtype=jnp.float32))
    assert jnp.all(info.terminated == flags)
    assert jnp.all(info.truncated == ~flags)


def test_state_slicing_matches_single_env_run():
    base = VectorToyEnv(dim=3)
    w = VmapWrapper(base, batch_size=4)
    key = jax.random.key(0)
    s_b, i_b = w.init(key)
    idx = 2
    # Run single env with the same per-env key used by the wrapper for index idx
    keys = jax.random.split(key, 4)
    s0, _ = base.init(keys[idx])
    a_single = base.action_space.sample(keys[idx])
    s1_single, i1_single = base.step(s0, a_single)
    # Run batched and slice
    act = w.action_space.sample(key)
    act = act.at[idx].set(a_single)
    s1_b, i1_b = w.step(s_b, act)
    assert jnp.allclose(i1_b.obs[idx], i1_single.obs)


# -----------------------------------------------------------------------------
# Core: Transform compatibility and serialization
# -----------------------------------------------------------------------------


def test_jit_compatibility_smoke():
    env = VectorToyEnv(dim=2)
    w = VmapWrapper(env, batch_size=3)
    key = jax.random.key(0)

    @jax.jit
    def run_once(k, a):
        s, _ = w.init(k)
        ns, inf = w.step(s, a)
        return ns, inf.obs

    # Sample outside jit to avoid tracing space construction in __post_init__
    act = w.action_space.sample(key)
    ns, obs = run_once(key, act)
    assert ns is not None
    assert obs.shape == (3, 2)


def test_pickle_serialization_of_state_and_info():
    env = ScalarToyEnv()
    w = VmapWrapper(env, batch_size=2)
    key = jax.random.key(0)
    state, info = w.init(key)
    payload = (state, info)
    blob = pickle.dumps(payload)
    loaded_state, loaded_info = pickle.loads(blob)
    # Basic equality on contents
    assert jnp.allclose(loaded_state, state)
    assert jnp.allclose(loaded_info.obs, info.obs)


# -----------------------------------------------------------------------------
# Optional: Property-based sampling
# -----------------------------------------------------------------------------


@given(
    batch_size=st.integers(min_value=1, max_value=4),
    dim=st.integers(min_value=1, max_value=5),
    seed=st.integers(min_value=0, max_value=2**31 - 1),
)
@settings(deadline=None, max_examples=20)
def test_prop_shapes_and_contains(batch_size, dim, seed):
    env = VectorToyEnv(dim=dim)
    w = VmapWrapper(env, batch_size=batch_size)
    key = jax.random.key(seed)
    state, info = w.init(key)
    assert state is not None
    assert isinstance(info, Info)
    assert w.observation_space.contains(info.obs)
    action = w.action_space.sample(key)
    assert w.action_space.contains(action)
    next_state, next_info = w.step(state, action)
    assert w.observation_space.contains(next_info.obs)



from abc import ABC, abstractmethod
from functools import cached_property
from typing import override

import jax
from jax import numpy as jnp

from envelope.struct import FrozenPyTreeNode, static_field
from envelope.typing import Key, PyTree


class Space(ABC, FrozenPyTreeNode):
    """Base class for all spaces. Spaces are immutable and hashable."""

    @property
    @abstractmethod
    def shape(self) -> tuple[int, ...] | PyTree:
        """The shape of the space."""

    @property
    @abstractmethod
    def dtype(self) -> jnp.dtype | PyTree:
        """The dtype of the space."""

    @abstractmethod
    def sample(self, key: Key) -> PyTree:
        """Sample a random element from the space."""

    @abstractmethod
    def contains(self, x: PyTree) -> bool:
        """Check if `self` contains a sample `x`."""


class Discrete(Space):
    """
    A discrete space with a given number of elements. `n` can be a scalar or an array.
    The shape and dtype of the space are inferred from `n`.

    Args:
        n (int | jax.Array): The number of elements in the space.
    """

    n: int | jax.Array

    @classmethod
    def from_shape(cls, n: int, shape: tuple[int, ...]) -> "Discrete":
        """
        Create a Discrete space from a shape and a number of elements. This is
        a shorthand for `Discrete` with `n` being expanded to the shape.
        """
        return cls(n=jnp.full(shape, n, dtype=jnp.asarray(n).dtype))

    @property
    def shape(self) -> tuple[int, ...]:
        return jnp.asarray(self.n).shape

    @property
    def dtype(self) -> jnp.dtype:
        return jnp.asarray(self.n).dtype

    def sample(self, key: Key) -> jax.Array:
        return jax.random.randint(key, self.shape, 0, self.n, dtype=self.dtype)

    def contains(self, x: int | jax.Array) -> bool:
        return jnp.all(x >= 0) & jnp.all(x < self.n)

    def __repr__(self) -> str:
        return f"Discrete(shape={self.shape}, dtype={self.dtype}, n={self.n})"


class Continuous(Space):
    """
    A continuous space with a given lower and upper bound. `low` and `high` can be
    scalars or arrays. The shape and dtype of the space are inferred from `low` and
    `high`.

    Args:
        low (float | jax.Array): The lower bound of the space.
        high (float | jax.Array): The upper bound of the space.
    """

    low: float | jax.Array
    high: float | jax.Array

    @classmethod
    def from_shape(
        cls, low: float, high: float, shape: tuple[int, ...]
    ) -> "Continuous":
        """
        Create a Continuous space from a shape and a lower and upper bound. This is a
        shorthand for `Continuous` with `low` and `high` being expanded to the shape.
        """
        return cls(
            low=jnp.full(shape, low, dtype=jnp.asarray(low).dtype),
            high=jnp.full(shape, high, dtype=jnp.asarray(high).dtype),
        )

    @property
    def dtype(self) -> jnp.dtype:
        if jnp.asarray(self.low).dtype != jnp.asarray(self.high).dtype:
            raise ValueError("low and high must have the same dtype")

        return jnp.asarray(self.low).dtype

    @property
    def shape(self) -> tuple[int, ...]:
        if jnp.asarray(self.low).shape != jnp.asarray(self.high).shape:
            raise ValueError("low and high must have the same shape")

        return jnp.asarray(self.low).shape

    @override
    def sample(self, key: Key) -> jax.Array:
        uniform_sample = jax.random.uniform(key, self.shape, self.dtype)
        return self.low + uniform_sample * (self.high - self.low)

    @override
    def contains(self, x: jax.Array) -> bool:
        return jnp.all((x >= jnp.asarray(self.low)) & (x <= jnp.asarray(self.high)))

    def __repr__(self) -> str:
        dtype_str = getattr(self.dtype, "__name__", str(self.dtype))
        return (
            f"Continuous(shape={self.shape}, dtype={dtype_str}, "
            f"low={self.low}, high={self.high})"
        )


class PyTreeSpace(Space):
    """
    A Space defined by a PyTree structure of other Spaces. While `PyTreeSpace.tree`
    might be an aribrarily nested PyTree, it's leaves must only be `Discrete` or
    `Continuous`, and not contain `PyTreeSpace` or `BatchedSpace` as leaves. The shape
    and dtype of the `PyTreeSpace` are PyTrees of the same structure, containing the
    shape and dtype of the leaves.

    Args:
        tree (PyTree): A PyTree with `Discrete` or `Continuous` leaves.
    """

    tree: PyTree

    def __post_init__(self):
        leaves = jax.tree.leaves(self.tree, is_leaf=lambda x: isinstance(x, Space))
        for leaf in leaves:
            if not isinstance(leaf, (Discrete, Continuous)):
                raise TypeError(
                    f"PyTreeSpace leaves must be Discrete or Continuous,"
                    f"got {type(leaf).__name__}"
                )

    @override
    def sample(self, key: Key) -> PyTree:
        leaves, treedef = jax.tree.flatten(
            self.tree, is_leaf=lambda x: isinstance(x, Space)
        )
        keys = jax.random.split(key, len(leaves))
        samples = [space.sample(key) for key, space in zip(keys, leaves)]
        return jax.tree.unflatten(treedef, samples)

    @override
    def contains(self, x: PyTree) -> bool:
        # Use tree.map to check containment for each space-value pair
        contains = jax.tree.map(
            lambda space, xi: space.contains(xi),
            self.tree,
            x,
            is_leaf=lambda node: isinstance(node, Space),
        )
        return jnp.all(jnp.array(jax.tree.leaves(contains)))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.tree!r})"

    @property
    def shape(self) -> PyTree:
        """A PyTree of the same structure as `tree`, containing each leaf's shape."""
        return jax.tree.map(
            lambda space: space.shape,
            self.tree,
            is_leaf=lambda node: isinstance(node, Space),
        )

    @property
    def dtype(self) -> PyTree:
        """A PyTree of the same structure as `tree`, containing each leaf's dtype."""
        return jax.tree.map(
            lambda space: space.dtype,
            self.tree,
            is_leaf=lambda node: isinstance(node, Space),
        )


def peel_batched(space: Space) -> tuple[tuple[int, ...], Space]:
    """Collect batch dimensions and return (batch_dims_tuple, base_space)."""
    dims: list[int] = []
    s: Space = space
    while isinstance(s, BatchedSpace):
        dims.append(s.batch_size)
        s = s.space
    return tuple(dims), s


class BatchedSpace(Space):
    """
    A view that adds a leading batch dimension to a base `Space` without
    materializing or broadcasting its parameters.

    Args:
        space (Space): The underlying base space.
        batch_size (int): The leading batch dimension.
    """

    space: Space
    batch_size: int = static_field()

    @override
    def sample(self, key: Key) -> PyTree:
        """
        Sample a batch of samples from the wrapped `Space`. You may pass a single key
        or a batch of keys shaped `(batch_size, ...)`.
        """
        if not jnp.issubdtype(key.dtype, jax.dtypes.prng_key):
            raise ValueError("key must be a (new-style) `jax.random.key`.")

        # Accept single key or a batch of keys shaped (batch_size, )
        if key.shape == ():
            keys = jax.random.split(key, self.batch_size)
        elif key.shape[0] == self.batch_size:
            keys = key
        else:
            raise ValueError(
                f"sample key's leading dimension ({key.shape[0]}) must match "
                f"batch_size ({self.batch_size})."
            )
        return jax.vmap(self.space.sample)(keys)

    @override
    def contains(self, x: PyTree) -> bool:
        """
        `BatchedSpace.contains` checks if each entry of `x` along the leading dimension
        is contained in the base (unbatched) `Space`.
        """
        result = jax.vmap(self.space.contains)(x)
        return jnp.all(jnp.asarray(result))

    @override
    @cached_property
    def shape(self) -> tuple[int, ...] | PyTree:
        """
        The shape of the `BatchedSpace` is the leading batch dimension prepend the
        shape of the wrapped `Space`. If the wrapped `Space` is a `PyTreeSpace`, the
        shape is a PyTree of the same structure, with the leading batch dimension
        prepended to the shape of each leaf `Space`.
        """
        batch_dims, base = peel_batched(self)
        if isinstance(base, PyTreeSpace):
            return jax.tree.map(
                lambda space: batch_dims + space.shape,
                base.tree,
                is_leaf=lambda node: isinstance(node, Space),
            )
        return batch_dims + base.shape

    @override
    @property
    def dtype(self) -> jnp.dtype | PyTree:
        """The dtype of the base space (batch dimensions don't affect dtype)."""
        _, base = peel_batched(self)
        return base.dtype

    def __repr__(self) -> str:
        return f"BatchedSpace(space={self.space!r}, batch_size={self.batch_size})"



import jax
import jax.numpy as jnp

from envelope.spaces import Continuous, PyTreeSpace
from envelope.wrappers.clip_action_wrapper import ClipActionWrapper, clip_action
from envelope.wrappers.vmap_wrapper import VmapWrapper
from tests.wrappers.helpers import (
    DiscreteStepCounterEnv,
    ScalarToyEnv,
    StepCounterEnv,
    VectorToyEnv,
)


def test_init_reset_delegate_unchanged():
    env = ScalarToyEnv()
    w = ClipActionWrapper(env)
    key = jax.random.key(0)
    state_w, info_w = w.init(key)
    state_e, info_e = env.init(key)
    assert jnp.allclose(state_w, state_e)
    assert jnp.allclose(info_w.obs, info_e.obs)

    state_w, info_w = w.reset(state_w, key)
    state_e, info_e = env.reset(state_e, key)
    assert jnp.allclose(state_w, state_e)
    assert jnp.allclose(info_w.obs, info_e.obs)
    assert w.observation_space.contains(info_w.obs)


def test_step_clips_continuous_scalar_action():
    env = StepCounterEnv()
    w = ClipActionWrapper(env)
    key = jax.random.key(0)
    state, _ = w.init(key)
    # Action 5.0 out of bounds [-1, 1] -> clipped to 1.0
    out_of_bounds = jnp.array(5.0)
    state_clipped, info_clipped = w.step(state, out_of_bounds)
    state_expected, info_expected = env.step(state, jnp.array(1.0))
    assert jnp.allclose(state_clipped.env_state, state_expected.env_state)
    assert jnp.allclose(state_clipped.steps, state_expected.steps)
    assert jnp.allclose(info_clipped.obs, info_expected.obs)


def test_step_clips_continuous_vector_action():
    env = VectorToyEnv(dim=3)
    w = ClipActionWrapper(env)
    key = jax.random.key(0)
    state, _ = w.init(key)
    # [2, -3, 0.5] -> [1, -1, 0.5] for bounds [-1, 1]
    action = jnp.array([2.0, -3.0, 0.5])
    state_w, info_w = w.step(state, action)
    clipped = jnp.array([1.0, -1.0, 0.5])
    state_e, info_e = env.step(state, clipped)
    assert jnp.allclose(state_w, state_e)
    assert jnp.allclose(info_w.obs, info_e.obs)


def test_in_bounds_action_passes_through_unchanged():
    env = ScalarToyEnv()
    w = ClipActionWrapper(env)
    key = jax.random.key(0)
    state, _ = w.init(key)
    action = jnp.array(0.3)
    state_w, info_w = w.step(state, action)
    state_e, info_e = env.step(state, action)
    assert jnp.allclose(state_w, state_e)
    assert jnp.allclose(info_w.obs, info_e.obs)


def test_clips_discrete_action():
    env = DiscreteStepCounterEnv(action_n=5)
    w = ClipActionWrapper(env)
    key = jax.random.key(0)
    state, _ = w.init(key)
    # action=10 with Discrete(5) -> valid [0,4], clip to 4
    action = jnp.array(10, dtype=jnp.int32)
    state_w, info_w = w.step(state, action)
    state_e, info_e = env.step(state, jnp.array(4, dtype=jnp.int32))
    assert jnp.allclose(state_w.env_state, state_e.env_state)
    assert jnp.allclose(info_w.obs, info_e.obs)


def test_clips_negative_discrete_action():
    env = DiscreteStepCounterEnv(action_n=5)
    w = ClipActionWrapper(env)
    key = jax.random.key(0)
    state, _ = w.init(key)
    action = jnp.array(-3, dtype=jnp.int32)
    state_w, info_w = w.step(state, action)
    state_e, info_e = env.step(state, jnp.array(0, dtype=jnp.int32))
    assert jnp.allclose(state_w.env_state, state_e.env_state)
    assert jnp.allclose(info_w.obs, info_e.obs)


def test_action_at_exact_boundary_passes_through():
    env = ScalarToyEnv()
    w = ClipActionWrapper(env)
    key = jax.random.key(0)
    state, _ = w.init(key)
    for bound in [jnp.array(-1.0), jnp.array(1.0)]:
        state_w, info_w = w.step(state, bound)
        state_e, info_e = env.step(state, bound)
        assert jnp.allclose(state_w, state_e)
        assert jnp.allclose(info_w.obs, info_e.obs)


def test_inf_bounds_env_large_action_passes_through():
    """When action space has inf bounds, clip_action leaves large actions unchanged. Test the function directly (no helper env has inf action bounds)."""
    space = Continuous(low=-jnp.inf, high=jnp.inf)
    action = jnp.array(1000.0)
    result = clip_action(action, space)
    assert jnp.allclose(result, action)


def test_spaces_unchanged():
    env = ScalarToyEnv()
    w = ClipActionWrapper(env)
    assert w.observation_space is env.observation_space
    assert w.action_space is env.action_space


def test_observation_space_contains_after_init_and_step():
    env = ScalarToyEnv()
    w = ClipActionWrapper(env)
    key = jax.random.key(0)
    state, info = w.init(key)
    assert w.observation_space.contains(info.obs)
    state, info = w.step(state, jnp.array(0.5))
    assert w.observation_space.contains(info.obs)


def test_jit_compatibility():
    env = ScalarToyEnv()
    w = ClipActionWrapper(env)
    key = jax.random.key(0)

    @jax.jit
    def step_jit(s, a):
        return w.step(s, a)

    state, _ = w.init(key)
    action = jnp.array(0.7)
    next_state, info = step_jit(state, action)
    assert jnp.allclose(next_state, state + action)
    assert jnp.allclose(info.obs, state + action)


def test_composability_with_vmap_wrapper():
    batch_size = 2
    env = ScalarToyEnv()
    w = ClipActionWrapper(VmapWrapper(env, batch_size=batch_size))
    key = jax.random.key(0)
    state, _ = w.init(key)
    # Batched out-of-bounds actions [5, -5] -> clipped to [1, -1]
    action = jnp.array([5.0, -5.0])
    next_state, info = w.step(state, action)
    assert jnp.allclose(info.obs, jnp.array([1.0, -1.0]))


def test_clip_action_function_with_pytree_space():
    space = PyTreeSpace(
        {
            "a": Continuous.from_shape(-1.0, 1.0, (2,)),
            "b": Continuous.from_shape(-1.0, 1.0, (3,)),
        }
    )
    action = {
        "a": jnp.array([2.0, -2.0]),
        "b": jnp.array([0.5, 10.0, -10.0]),
    }
    result = clip_action(action, space)
    assert jnp.allclose(result["a"], jnp.array([1.0, -1.0]))
    assert jnp.allclose(result["b"], jnp.array([0.5, 1.0, -1.0]))



from typing import override

import jax.numpy as jnp

from envelope.environment import Info
from envelope.struct import field
from envelope.typing import Key, PyTree
from envelope.wrappers.wrapper import WrappedState, Wrapper


class TruncationWrapper(Wrapper):
    max_steps: int = field(kw_only=True)

    class TruncationState(WrappedState):
        steps: jnp.ndarray | int = field(default=0)

    @override
    def init(self, key: Key) -> tuple[WrappedState, Info]:
        inner_state, info = self.env.init(key)
        state = self.TruncationState(inner_state=inner_state, steps=0)
        return state, info.update(truncated=self.max_steps <= 0)

    @override
    def reset(self, state: WrappedState, key: Key) -> tuple[WrappedState, Info]:
        inner_state, info = self.env.reset(state.inner_state, key)
        state = state.replace(inner_state=inner_state, steps=0)
        return state, info.update(truncated=self.max_steps <= 0)

    @override
    def step(self, state: WrappedState, action: PyTree) -> tuple[WrappedState, Info]:
        next_inner_state, info = self.env.step(state.inner_state, action)
        steps = state.steps + 1
        state = self.TruncationState(inner_state=next_inner_state, steps=steps)
        truncated = jnp.asarray(steps) >= self.max_steps
        return state, info.update(truncated=truncated)



from functools import cached_property
from typing import override

import jax
import jax.numpy as jnp

from envelope import spaces
from envelope.environment import Environment, Info
from envelope.struct import field
from envelope.typing import Key, PyTree
from envelope.wrappers.wrapper import WrappedState, Wrapper


class VmapEnvsWrapper(Wrapper):
    """
    Vectorizes over a batched collection of environment instances (vmapping over 'self').

    Usage:
        ```python
        envs = jax.vmap(make_env)(params_batch)  # env pytree batched on leading axis
        wrapped = VmapEnvsWrapper(env=envs, batch_size=B)
        state, info = wrapped.init(keys)  # keys shape (B, 2) or single key
        next_state, info = wrapped.step(state, action)
        ```
    """

    batch_size: int = field(kw_only=True)

    def _split_keys(self, key: Key) -> Key:
        if not jnp.issubdtype(key.dtype, jax.dtypes.prng_key):
            raise ValueError("key must be a (new-style) `jax.random.key`.")

        if key.shape == ():
            return jax.random.split(key, self.batch_size)
        elif key.shape[0] == self.batch_size:
            return key
        raise ValueError(
            f"reset key's leading dimension ({key.shape[0]}) must match "
            f"batch_size ({self.batch_size})."
        )

    @override
    def init(self, key: Key) -> tuple[WrappedState, Info]:
        keys = self._split_keys(key)
        state, info = jax.vmap(lambda e, k: e.init(k))(self.env, keys)
        return state, info

    @override
    def reset(self, state: PyTree, key: Key) -> tuple[WrappedState, Info]:
        keys = self._split_keys(key)
        state, info = jax.vmap(lambda e, s, k: e.reset(s, k))(self.env, state, keys)
        return state, info

    @override
    def step(self, state: WrappedState, action: PyTree) -> tuple[WrappedState, Info]:
        next_state, info = jax.vmap(lambda e, s, a: e.step(s, a))(
            self.env, state, action
        )
        return next_state, info

    @override
    @property
    def observation_space(self) -> spaces.Space:
        env0 = _index_env(self.env, 0, self.batch_size)
        return spaces.BatchedSpace(
            space=env0.observation_space, batch_size=self.batch_size
        )

    @override
    @cached_property
    def action_space(self) -> spaces.Space:
        env0 = _index_env(self.env, 0, self.batch_size)
        return spaces.BatchedSpace(space=env0.action_space, batch_size=self.batch_size)

    @override
    @property
    def unwrapped(self) -> Environment:
        return self.env.unwrapped


def _index_env(env: Environment, idx: int, batch_size: int) -> Environment:
    def idx_or_keep(x):
        if hasattr(x, "shape") and isinstance(getattr(x, "shape"), tuple):
            if len(x.shape) > 0 and x.shape[0] == batch_size:
                return x[idx]
        return x

    return jax.tree.map(lambda x: idx_or_keep(x), env)



from functools import cached_property
from typing import override

import jax
import jax.numpy as jnp

from envelope.environment import Info, State
from envelope.spaces import BatchedSpace, Continuous, Discrete, Space, peel_batched
from envelope.typing import Key, PyTree
from envelope.wrappers.wrapper import Wrapper


def to_float(obs: PyTree) -> PyTree:
    return jax.tree.map(lambda x: x.astype(jnp.float32), obs)


def to_continuous(space: Discrete | Continuous) -> Continuous:
    if isinstance(space, Continuous):
        low = jnp.asarray(space.low, dtype=jnp.float32)
        high = jnp.asarray(space.high, dtype=jnp.float32)
        return Continuous(low=low, high=high)
    elif isinstance(space, Discrete):
        n = jnp.asarray(space.n)
        low = jnp.zeros_like(n, dtype=jnp.float32)
        high = jnp.asarray(n - 1, dtype=jnp.float32)
        return Continuous(low=low, high=high)
    raise TypeError(f"Expected Discrete or Continuous, got {type(space)}")


class ContinuousObservationWrapper(Wrapper):
    @override
    def init(self, key: Key) -> tuple[State, Info]:
        state, info = self.env.init(key)
        info = info.update(obs=to_float(info.obs))
        return state, info

    @override
    def reset(self, state: State, key: Key) -> tuple[State, Info]:
        state, info = self.env.reset(state, key)
        info = info.update(obs=to_float(info.obs))
        return state, info

    @override
    def step(self, state: State, action: PyTree) -> tuple[State, Info]:
        state, info = self.env.step(state, action)
        info = info.update(obs=to_float(info.obs))
        return state, info

    @override
    @cached_property
    def observation_space(self) -> Space:
        batch_dims, base = peel_batched(self.env.observation_space)

        def is_leaf(x):
            return isinstance(x, (Discrete, Continuous))

        space = jax.tree.map(to_continuous, base, is_leaf=is_leaf)

        for batch_dim in batch_dims:
            space = BatchedSpace(space, batch_dim)
        return space



from typing import override

import jax

from envelope.environment import Info, State
from envelope.struct import FrozenPyTreeNode, field
from envelope.typing import Key, PyTree
from envelope.wrappers.wrapper import WrappedState, Wrapper


class EpisodeStatistics(FrozenPyTreeNode):
    reward: jax.Array = field(default=0.0)
    length: jax.Array = field(default=0.0)


class EpisodeStatisticsWrapper(Wrapper):
    class EpisodeStatisticsState(WrappedState):
        stats: EpisodeStatistics = field(default=EpisodeStatistics())

    @override
    def init(self, key: Key) -> tuple[State, Info]:
        inner_state, info = self.env.init(key)
        state = self.EpisodeStatisticsState(inner_state=inner_state)
        return state, info.update(stats=state.stats)

    @override
    def reset(self, state: State, key: Key) -> tuple[State, Info]:
        inner_state, info = self.env.reset(state.inner_state, key)
        state = state.replace(inner_state=inner_state)
        return state, info.update(stats=state.stats)

    @override
    def step(self, state: State, action: PyTree) -> tuple[State, Info]:
        inner_state, info = self.env.step(state.inner_state, action)
        stats = state.stats.replace(
            reward=state.stats.reward + info.reward,
            length=state.stats.length + 1,
        )
        state = state.replace(inner_state=inner_state, stats=stats)
        return state, info.update(stats=stats)



"""Tests for envelope.struct module.

This module tests:
- Immutability (frozen dataclasses)
- JAX pytree integration (tree_flatten/unflatten)
- JAX transformations (vmap, jit, tree.map)
- Dataclass features (defaults, inheritance, replace)
- Serialization (pickle, repr)
"""

import dataclasses

import jax
import jax.numpy as jnp
import pytest

from envelope.struct import FrozenPyTreeNode, field, static_field

# ============================================================================
# Test Fixtures and Helper Classes
# ============================================================================


class SimpleNode(FrozenPyTreeNode):
    """Test class with dynamic and static fields."""

    x: jax.Array  # dynamic field (pytree node)
    y: jax.Array  # dynamic field (pytree node)
    static_val: int = static_field()  # static field


class _NodeWithDefaults(FrozenPyTreeNode):
    """Helper node used to exercise default factories and static fields."""

    x: jax.Array
    y: jax.Array = field(default_factory=lambda: jnp.array([0.0]))
    scale: float = static_field(default=1.0)


class _NodeWithDefaultList(FrozenPyTreeNode):
    """Helper node for default_factory independence checks."""

    x: jax.Array
    items: list = static_field(default_factory=list)


class _BaseNode(FrozenPyTreeNode):
    """Helper base node for inheritance tests."""

    x: jax.Array
    config: str = static_field(default="base")


class _DerivedNode(_BaseNode):
    """Helper derived node for inheritance tests."""

    y: jax.Array = field(default_factory=lambda: jnp.array([0.0]))
    extra: int = static_field(default=0)


class _OrderedNode(FrozenPyTreeNode):
    """Helper node for verifying flatten ordering and stability."""

    a: jax.Array
    b: jax.Array
    static_1: int = static_field()
    c: jax.Array
    static_2: str = static_field()
    d: jax.Array


class _ComplexNode(FrozenPyTreeNode):
    """Helper node for exercising round-trip with diverse static fields."""

    array_field: jax.Array
    int_static: int = static_field()
    another_array: jax.Array
    str_static: str = static_field()
    list_static: list = static_field()


# ============================================================================
# Tests: Basic PyTreeNode Functionality
# ============================================================================


class TestPyTreeNodeBasics:
    """Test basic PyTreeNode functionality: immutability, replace, equality."""

    def test_frozen_attributes(self):
        """Test that PyTreeNode instances are frozen (immutable)."""
        node = SimpleNode(
            x=jnp.array([1.0, 2.0]), y=jnp.array([3.0, 4.0]), static_val=42
        )

        # Attempting to modify any attribute should raise FrozenInstanceError
        with pytest.raises(dataclasses.FrozenInstanceError):
            node.x = jnp.array([5.0, 6.0])

        with pytest.raises(dataclasses.FrozenInstanceError):
            node.y = jnp.array([7.0, 8.0])

        with pytest.raises(dataclasses.FrozenInstanceError):
            node.static_val = 100

    def test_replace_method(self):
        """Test the replace method works correctly."""
        node = SimpleNode(
            x=jnp.array([1.0, 2.0]), y=jnp.array([3.0, 4.0]), static_val=10
        )

        # Replace a dynamic field
        new_node = node.replace(x=jnp.array([5.0, 6.0]))
        assert jnp.allclose(new_node.x, jnp.array([5.0, 6.0]))
        assert jnp.allclose(new_node.y, jnp.array([3.0, 4.0]))  # unchanged
        assert new_node.static_val == 10  # unchanged

        # Replace a static field
        new_node2 = node.replace(static_val=20)
        assert jnp.allclose(new_node2.x, jnp.array([1.0, 2.0]))  # unchanged
        assert jnp.allclose(new_node2.y, jnp.array([3.0, 4.0]))  # unchanged
        assert new_node2.static_val == 20

        # Replace multiple fields
        new_node3 = node.replace(
            x=jnp.array([7.0, 8.0]), y=jnp.array([9.0, 10.0]), static_val=30
        )
        assert jnp.allclose(new_node3.x, jnp.array([7.0, 8.0]))
        assert jnp.allclose(new_node3.y, jnp.array([9.0, 10.0]))
        assert new_node3.static_val == 30

        # Original should be unchanged
        assert jnp.allclose(node.x, jnp.array([1.0, 2.0]))
        assert jnp.allclose(node.y, jnp.array([3.0, 4.0]))
        assert node.static_val == 10

    def test_replace_with_invalid_field(self):
        """Test that replace raises an error for non-existent fields."""
        node = SimpleNode(
            x=jnp.array([1.0, 2.0]), y=jnp.array([3.0, 4.0]), static_val=10
        )

        # Attempting to replace a non-existent field should raise TypeError
        with pytest.raises(TypeError, match="unexpected keyword argument"):
            node.replace(nonexistent_field=42)

        with pytest.raises(TypeError, match="unexpected keyword argument"):
            node.replace(x=jnp.array([5.0, 6.0]), invalid_param="test")

    def test_equality(self):
        """Test that PyTreeNode instances can be compared for equality."""
        node1 = SimpleNode(
            x=jnp.array([1.0, 2.0]), y=jnp.array([3.0, 4.0]), static_val=10
        )
        node2 = SimpleNode(
            x=jnp.array([1.0, 2.0]), y=jnp.array([3.0, 4.0]), static_val=10
        )
        node3 = SimpleNode(
            x=jnp.array([1.0, 2.0]), y=jnp.array([3.0, 4.0]), static_val=20
        )

        # Use tree_leaves to compare the pytree structures element-by-element
        leaves1, treedef1 = jax.tree_util.tree_flatten(node1)
        leaves2, treedef2 = jax.tree_util.tree_flatten(node2)
        leaves3, treedef3 = jax.tree_util.tree_flatten(node3)

        # node1 and node2 should be equal (same static and dynamic values)
        assert treedef1 == treedef2
        assert all(jnp.allclose(l1, l2) for l1, l2 in zip(leaves1, leaves2))

        # node1 and node3 should not be equal (different static_val means different treedef)
        assert treedef1 != treedef3  # Different static values = different treedefs


# ============================================================================
# Tests: JAX Integration
# ============================================================================


class TestPyTreeNodeJAXIntegration:
    """Test JAX-specific operations: vmap, jit, tree operations."""

    def test_vmap_dynamic_attributes(self):
        """Test that vmap works correctly over dynamic (pytree node) attributes."""

        # Create a function that operates on SimpleNode
        def process(node: SimpleNode) -> jax.Array:
            return node.x + node.y

        # Create a batch of nodes
        x_batch = jnp.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]])
        y_batch = jnp.array([[10.0, 20.0], [30.0, 40.0], [50.0, 60.0]])

        # Create batched node (static_val should be the same for all)
        batched_node = SimpleNode(x=x_batch, y=y_batch, static_val=42)

        # vmap over the batched dimensions
        result = jax.vmap(process)(batched_node)

        # Expected result
        expected = jnp.array([[11.0, 22.0], [33.0, 44.0], [55.0, 66.0]])

        assert jnp.allclose(result, expected)

    def test_vmap_static_attributes(self):
        """Test that static attributes are broadcast correctly in vmap."""

        # Create a function that uses both dynamic and static fields
        def process(node: SimpleNode) -> jax.Array:
            # Static field should be the same across all batched elements
            return node.x * node.static_val + node.y

        # Create a batch of nodes
        x_batch = jnp.array([[1.0, 2.0], [3.0, 4.0]])
        y_batch = jnp.array([[10.0, 20.0], [30.0, 40.0]])

        batched_node = SimpleNode(x=x_batch, y=y_batch, static_val=5)

        # vmap should broadcast the static field across all batch elements
        result = jax.vmap(process)(batched_node)

        # Expected: x * 5 + y for each batch element
        expected = jnp.array([[15.0, 30.0], [45.0, 60.0]])

        assert jnp.allclose(result, expected)

    def test_tree_flatten_unflatten(self):
        """Test that tree_flatten and tree_unflatten work correctly."""
        node = SimpleNode(
            x=jnp.array([1.0, 2.0]), y=jnp.array([3.0, 4.0]), static_val=42
        )

        # Flatten
        children, treedef = jax.tree_util.tree_flatten(node)

        # children should contain dynamic fields (x, y)
        assert len(children) == 2
        assert jnp.allclose(children[0], jnp.array([1.0, 2.0]))
        assert jnp.allclose(children[1], jnp.array([3.0, 4.0]))

        # Unflatten
        reconstructed = jax.tree_util.tree_unflatten(treedef, children)

        assert jnp.allclose(reconstructed.x, node.x)
        assert jnp.allclose(reconstructed.y, node.y)
        assert reconstructed.static_val == node.static_val

    def test_jax_tree_map(self):
        """Test that jax.tree.map works correctly with PyTreeNode."""
        node = SimpleNode(
            x=jnp.array([1.0, 2.0]), y=jnp.array([3.0, 4.0]), static_val=10
        )

        # tree.map should only transform dynamic fields
        result = jax.tree.map(lambda x: x * 2, node)

        assert jnp.allclose(result.x, jnp.array([2.0, 4.0]))
        assert jnp.allclose(result.y, jnp.array([6.0, 8.0]))
        assert result.static_val == 10  # static field unchanged

    def test_pytree_node_with_jit(self):
        """Test that PyTreeNode works with jax.jit."""

        @jax.jit
        def process(node: SimpleNode) -> jax.Array:
            return node.x * node.static_val + node.y

        node = SimpleNode(
            x=jnp.array([1.0, 2.0]), y=jnp.array([3.0, 4.0]), static_val=2
        )
        result = process(node)

        expected = jnp.array([5.0, 8.0])  # [1*2+3, 2*2+4]
        assert jnp.allclose(result, expected)

    def test_tree_flatten_preserves_order_and_is_stable(self):
        """Ensure flatten ordering is deterministic and round-trips."""
        node = _OrderedNode(
            a=jnp.array([1.0]),
            b=jnp.array([2.0]),
            static_1=10,
            c=jnp.array([3.0]),
            static_2="test",
            d=jnp.array([4.0]),
        )

        children1, treedef1 = jax.tree_util.tree_flatten(node)
        children2, treedef2 = jax.tree_util.tree_flatten(node)

        expected_children = (
            jnp.array([1.0]),
            jnp.array([2.0]),
            jnp.array([3.0]),
            jnp.array([4.0]),
        )

        assert len(children1) == len(expected_children) == 4
        for child, expected in zip(children1, expected_children):
            assert jnp.allclose(child, expected)

        assert len(children1) == len(children2)
        assert all(jnp.allclose(c1, c2) for c1, c2 in zip(children1, children2))
        assert treedef1 == treedef2

        reconstructed = jax.tree_util.tree_unflatten(treedef1, children1)
        assert jnp.allclose(reconstructed.a, node.a)
        assert jnp.allclose(reconstructed.b, node.b)
        assert reconstructed.static_1 == node.static_1
        assert jnp.allclose(reconstructed.c, node.c)
        assert reconstructed.static_2 == node.static_2
        assert jnp.allclose(reconstructed.d, node.d)

    @pytest.mark.parametrize(
        "node",
        [
            pytest.param(
                SimpleNode(
                    x=jnp.array([1.0, 2.0, 3.0]),
                    y=jnp.array([4.0, 5.0, 6.0]),
                    static_val=99,
                ),
                id="simple-node",
            ),
            pytest.param(
                _ComplexNode(
                    array_field=jnp.array([1.0, 2.0, 3.0]),
                    int_static=42,
                    another_array=jnp.array([[1.0, 2.0], [3.0, 4.0]]),
                    str_static="metadata",
                    list_static=[1, 2, 3],
                ),
                id="complex-node",
            ),
        ],
    )
    def test_tree_flatten_round_trip(self, node):
        """Flatten/unflatten should be a no-op for representative nodes."""
        children, treedef = jax.tree_util.tree_flatten(node)
        reconstructed = jax.tree_util.tree_unflatten(treedef, children)

        for field_info in dataclasses.fields(node):
            original_value = getattr(node, field_info.name)
            reconstructed_value = getattr(reconstructed, field_info.name)
            if isinstance(original_value, jax.Array):
                assert jnp.allclose(reconstructed_value, original_value)
            else:
                assert reconstructed_value == original_value


# ============================================================================
# Tests: Nested Structures
# ============================================================================


class TestPyTreeNodeNesting:
    """Test nested PyTreeNode structures."""

    def test_nested_pytree_nodes(self):
        """Test nested PyTreeNode structures."""

        class Outer(FrozenPyTreeNode):
            inner: SimpleNode
            z: jax.Array

        inner_node = SimpleNode(x=jnp.array([1.0]), y=jnp.array([2.0]), static_val=5)
        outer_node = Outer(inner=inner_node, z=jnp.array([3.0]))

        # Test that nested structures work with tree.map
        result = jax.tree.map(lambda x: x * 10, outer_node)

        assert jnp.allclose(result.inner.x, jnp.array([10.0]))
        assert jnp.allclose(result.inner.y, jnp.array([20.0]))
        assert result.inner.static_val == 5  # static field unchanged
        assert jnp.allclose(result.z, jnp.array([30.0]))

    def test_nested_defaults(self):
        """Test nodes with defaults containing other nodes with defaults."""

        class InnerNode(FrozenPyTreeNode):
            value: jax.Array = field(default_factory=lambda: jnp.array([0.0]))

        class OuterNode(FrozenPyTreeNode):
            data: jax.Array
            inner: InnerNode = field(default_factory=InnerNode)

        # Create outer node with default inner
        outer = OuterNode(data=jnp.array([5.0]))
        assert jnp.allclose(outer.data, jnp.array([5.0]))
        assert jnp.allclose(outer.inner.value, jnp.array([0.0]))

        # Create outer node with custom inner
        custom_inner = InnerNode(value=jnp.array([10.0]))
        outer2 = OuterNode(data=jnp.array([15.0]), inner=custom_inner)
        assert jnp.allclose(outer2.data, jnp.array([15.0]))
        assert jnp.allclose(outer2.inner.value, jnp.array([10.0]))


# ============================================================================
# Tests: Default Values
# ============================================================================


class TestPyTreeNodeDefaults:
    """Test default values and default factories."""

    def test_node_with_defaults_behavior(self):
        """Ensure default factories and static fields behave as expected."""
        explicit = _NodeWithDefaults(
            x=jnp.array([1.0, 2.0]),
            y=jnp.array([3.0, 4.0]),
            scale=2.0,
        )
        assert jnp.allclose(explicit.x, jnp.array([1.0, 2.0]))
        assert jnp.allclose(explicit.y, jnp.array([3.0, 4.0]))
        assert explicit.scale == 2.0

        # Default values populate when parameters are omitted
        defaulted = _NodeWithDefaults(x=jnp.array([5.0, 6.0]))
        assert jnp.allclose(defaulted.y, jnp.array([0.0]))
        assert defaulted.scale == 1.0

        # tree.map should only transform dynamic fields
        transformed = jax.tree.map(lambda arr: arr * 10, explicit)
        assert jnp.allclose(transformed.x, jnp.array([10.0, 20.0]))
        assert jnp.allclose(transformed.y, jnp.array([30.0, 40.0]))
        assert transformed.scale == 2.0

        # tree_flatten / tree_unflatten should round-trip the node
        children, treedef = jax.tree_util.tree_flatten(explicit)
        assert len(children) == 2
        reconstructed = jax.tree_util.tree_unflatten(treedef, children)
        assert jnp.allclose(reconstructed.x, explicit.x)
        assert jnp.allclose(reconstructed.y, explicit.y)
        assert reconstructed.scale == explicit.scale

        # vmap should work when batching NodeWithDefaults instances
        def process(node: _NodeWithDefaults) -> jax.Array:
            return node.x * node.scale + node.y

        x_batch = jnp.array([[1.0], [2.0], [3.0]])
        y_batch = jnp.zeros_like(x_batch)
        batched_node = _NodeWithDefaults(x=x_batch, y=y_batch)
        result = jax.vmap(process)(batched_node)
        expected = jnp.array([[1.0], [2.0], [3.0]])
        assert jnp.allclose(result, expected)

    def test_static_field_with_default(self):
        """Test that static_field() supports default values."""

        class NodeWithStaticDefault(FrozenPyTreeNode):
            x: jax.Array
            config: dict = static_field(default_factory=dict)
            threshold: float = static_field(default=0.5)

        # Create with defaults
        node1 = NodeWithStaticDefault(x=jnp.array([1.0]))
        assert jnp.allclose(node1.x, jnp.array([1.0]))
        assert node1.config == {}
        assert node1.threshold == 0.5

        # Create with explicit values
        node2 = NodeWithStaticDefault(
            x=jnp.array([2.0]), config={"key": "value"}, threshold=0.9
        )
        assert jnp.allclose(node2.x, jnp.array([2.0]))
        assert node2.config == {"key": "value"}
        assert node2.threshold == 0.9

    def test_mixed_required_and_default_fields(self):
        """Test nodes with mix of required and default fields."""

        class NodeMixedDefaults(FrozenPyTreeNode):
            # Required fields
            a: jax.Array
            b: int = static_field()

            # Optional fields with defaults
            c: jax.Array = field(default_factory=lambda: jnp.array([1.0]))
            d: str = static_field(default="default_value")
            e: jax.Array = field(default_factory=lambda: jnp.array([2.0, 3.0]))

        # Provide only required fields
        node1 = NodeMixedDefaults(a=jnp.array([10.0]), b=5)
        assert jnp.allclose(node1.a, jnp.array([10.0]))
        assert node1.b == 5
        assert jnp.allclose(node1.c, jnp.array([1.0]))
        assert node1.d == "default_value"
        assert jnp.allclose(node1.e, jnp.array([2.0, 3.0]))

        # Override some defaults
        node2 = NodeMixedDefaults(
            a=jnp.array([20.0]), b=10, c=jnp.array([5.0, 6.0]), d="custom"
        )
        assert jnp.allclose(node2.a, jnp.array([20.0]))
        assert node2.b == 10
        assert jnp.allclose(node2.c, jnp.array([5.0, 6.0]))
        assert node2.d == "custom"
        assert jnp.allclose(node2.e, jnp.array([2.0, 3.0]))

    def test_default_factory_independence(self):
        """Test that default_factory creates independent instances."""

        node1 = _NodeWithDefaultList(x=jnp.array([1.0]))
        node2 = _NodeWithDefaultList(x=jnp.array([2.0]))

        # Modify node1's list (if it weren't frozen)
        # Since the node is frozen, we can't modify in place, but we can check
        # that the default_factory created different instances
        assert node1.items is not node2.items
        assert node1.items == []
        assert node2.items == []


# ============================================================================
# Tests: Inheritance
# ============================================================================


class TestPyTreeNodeInheritance:
    """Test inheritance behavior of PyTreeNode."""

    def test_pytree_inheritance_core_behaviour(self):
        """Ensure inherited PyTreeNodes support core dataclass behaviours."""
        node = _DerivedNode(
            x=jnp.array([1.0, 2.0]),
            y=jnp.array([3.0, 4.0]),
            config="derived",
            extra=42,
        )

        assert jnp.allclose(node.x, jnp.array([1.0, 2.0]))
        assert jnp.allclose(node.y, jnp.array([3.0, 4.0]))
        assert node.config == "derived"
        assert node.extra == 42

        with pytest.raises(dataclasses.FrozenInstanceError):
            node.x = jnp.array([5.0])

        children, treedef = jax.tree_util.tree_flatten(node)
        assert len(children) == 2  # dynamic fields

        mapped = jax.tree.map(lambda arr: arr * 2, node)
        assert jnp.allclose(mapped.x, jnp.array([2.0, 4.0]))
        assert jnp.allclose(mapped.y, jnp.array([6.0, 8.0]))
        assert mapped.config == "derived"
        assert mapped.extra == 42

        replaced = node.replace(x=jnp.array([10.0, 20.0]), config="updated")
        assert jnp.allclose(replaced.x, jnp.array([10.0, 20.0]))
        assert jnp.allclose(replaced.y, jnp.array([3.0, 4.0]))
        assert replaced.config == "updated"
        assert replaced.extra == 42

    def test_pytree_inheritance_vmap_and_defaults(self):
        """Check vectorisation and default handling for inherited nodes."""

        node_with_defaults = _DerivedNode(x=jnp.array([5.0]))
        assert jnp.allclose(node_with_defaults.y, jnp.array([0.0]))
        assert node_with_defaults.config == "base"
        assert node_with_defaults.extra == 0

        def process(node: _DerivedNode) -> jax.Array:
            return node.x + node.y

        x_batch = jnp.array([[1.0, 2.0], [3.0, 4.0]])
        y_batch = jnp.array([[10.0, 20.0], [30.0, 40.0]])
        batched = _DerivedNode(x=x_batch, y=y_batch, config="batch", extra=99)

        result = jax.vmap(process)(batched)
        expected = jnp.array([[11.0, 22.0], [33.0, 44.0]])
        assert jnp.allclose(result, expected)
        assert batched.config == "batch"
        assert batched.extra == 99


# ============================================================================
# Tests: Serialization
# ============================================================================


class TestPyTreeNodeSerialization:
    """Test serialization: repr, hashing, and pickling."""

    def test_pytree_node_repr(self):
        """Test that PyTreeNode has a readable __repr__."""
        node = SimpleNode(
            x=jnp.array([1.0, 2.0]), y=jnp.array([3.0, 4.0]), static_val=42
        )

        repr_str = repr(node)

        # Should contain class name
        assert "SimpleNode" in repr_str

        # Should be a valid representation (dataclass default repr)
        # Check that it contains field names
        assert "x=" in repr_str or "static_val=" in repr_str

    def test_pytree_node_repr_with_defaults(self):
        """Test __repr__ with nodes that have default values."""

        class NodeWithDefaults(FrozenPyTreeNode):
            x: jax.Array
            y: jax.Array = field(default_factory=lambda: jnp.array([0.0]))
            meta: str = static_field(default="default_meta")

        node = NodeWithDefaults(x=jnp.array([1.0]))
        repr_str = repr(node)

        assert "NodeWithDefaults" in repr_str
        assert "x=" in repr_str

    def test_pytree_node_repr_nested(self):
        """Test __repr__ with nested PyTreeNode structures."""

        class Inner(FrozenPyTreeNode):
            value: jax.Array

        class Outer(FrozenPyTreeNode):
            inner: Inner
            data: jax.Array

        inner = Inner(value=jnp.array([1.0]))
        outer = Outer(inner=inner, data=jnp.array([2.0]))

        repr_str = repr(outer)

        # Should contain both class names
        assert "Outer" in repr_str
        assert "Inner" in repr_str

    def test_pickle_round_trip_basic(self):
        """Basic pickle/unpickle should preserve fields."""
        import pickle

        node = SimpleNode(
            x=jnp.array([1.0, 2.0]), y=jnp.array([3.0, 4.0]), static_val=42
        )

        # Pickle and unpickle
        pickled = pickle.dumps(node)
        unpickled = pickle.loads(pickled)

        # Verify fields are preserved
        assert jnp.allclose(unpickled.x, node.x)
        assert jnp.allclose(unpickled.y, node.y)
        assert unpickled.static_val == node.static_val

    def test_pickle_round_trip_comprehensive(self):
        """Pickling with varied field types; also verify treedef equality.

        Note: We use SimpleNode (module-level class) because local classes
        cannot be pickled in Python.
        """
        import pickle

        # Test with different values
        node = SimpleNode(
            x=jnp.array([1.0, 2.0, 3.0]),
            y=jnp.array([[1.0, 2.0], [3.0, 4.0]]),
            static_val=999,
        )

        # Pickle and unpickle
        pickled = pickle.dumps(node)
        unpickled = pickle.loads(pickled)

        # Verify all fields are preserved
        assert jnp.allclose(unpickled.x, node.x)
        assert jnp.allclose(unpickled.y, node.y)
        assert unpickled.static_val == node.static_val

        # Verify functionality still works
        new_node = unpickled.replace(static_val=1000)
        assert new_node.static_val == 1000
        assert jnp.allclose(new_node.x, node.x)

        # Also verify treedef/structure equality post-pickle
        leaves1, treedef1 = jax.tree_util.tree_flatten(node)
        leaves2, treedef2 = jax.tree_util.tree_flatten(unpickled)
        assert treedef1 == treedef2
        assert all(jnp.allclose(l1, l2) for l1, l2 in zip(leaves1, leaves2))

    def test_pytree_node_not_hashable(self):
        """Test that PyTreeNode instances are not hashable (because they're not).

        Frozen dataclasses with mutable fields (like jax.Array) are not hashable.
        This test documents the expected behavior.
        """
        node = SimpleNode(
            x=jnp.array([1.0, 2.0]), y=jnp.array([3.0, 4.0]), static_val=42
        )

        # PyTreeNode with array fields should not be hashable
        with pytest.raises(TypeError):
            hash(node)

        # Also test that we can't use it as a dict key
        with pytest.raises(TypeError):
            {node: "value"}



import pickle

import jax
import jax.numpy as jnp
import pytest

from envelope.environment import Info
from envelope.spaces import BatchedSpace
from envelope.wrappers.vmap_envs_wrapper import VmapEnvsWrapper
from tests.wrappers.helpers import ParamEnv

# -----------------------------------------------------------------------------
# Core: Space shaping and protocol conformance
# -----------------------------------------------------------------------------


@pytest.mark.parametrize("batch_size", [1, 3])
def test_spaces_are_batched(batch_size):
    # Build a batched pytree of envs via vmap(make_env)(params)
    params = jnp.linspace(-1.0, 1.0, batch_size)
    envs = ParamEnv(offset=params)
    w = VmapEnvsWrapper(envs, batch_size=batch_size)
    assert isinstance(w.observation_space, BatchedSpace)
    assert isinstance(w.action_space, BatchedSpace)
    assert w.observation_space.shape == (batch_size,) + (2,)
    assert w.action_space.shape == (batch_size,) + (2,)


def test_protocol_conformance_reset_and_step():
    batch_size = 4
    params = jnp.linspace(0.0, 0.3, batch_size)
    envs = ParamEnv(offset=params)
    w = VmapEnvsWrapper(envs, batch_size=batch_size)
    key = jax.random.key(0)
    state, info = w.init(key)
    assert state is not None
    assert isinstance(info, Info)
    action = w.action_space.sample(key)
    next_state, next_info = w.step(state, action)
    assert next_state is not None
    assert isinstance(next_info, Info)


# -----------------------------------------------------------------------------
# Core: Equivalence vs manual vmap over envs and keys
# -----------------------------------------------------------------------------


@pytest.mark.parametrize("batch_size", [2, 5])
def test_reset_and_step_equivalence_to_manual(batch_size):
    params = jnp.linspace(-0.5, 0.5, batch_size)
    envs = ParamEnv(offset=params)
    w = VmapEnvsWrapper(envs, batch_size=batch_size)
    key = jax.random.key(0)
    keys = jax.random.split(key, batch_size)
    s_w, i_w = w.init(key)
    # Manual: vmap over (env, key)
    s_m, i_m = jax.vmap(lambda e, k: e.init(k))(envs, keys)
    assert jax.tree_util.tree_all(
        jax.tree.map(lambda a, b: jnp.allclose(a, b), s_w, s_m)
    )
    assert jnp.allclose(i_w.obs, i_m.obs)
    action = w.action_space.sample(key)
    s_w2, i_w2 = w.step(s_w, action)
    s_m2, i_m2 = jax.vmap(lambda e, s, a: e.step(s, a))(envs, s_m, action)
    assert jax.tree_util.tree_all(
        jax.tree.map(lambda a, b: jnp.allclose(a, b), s_w2, s_m2)
    )
    assert jnp.allclose(i_w2.obs, i_m2.obs)
    assert jnp.allclose(i_w2.reward, i_m2.reward)


# -----------------------------------------------------------------------------
# Core: Error paths and serialization
# -----------------------------------------------------------------------------


def test_reset_raises_on_wrong_key_batch_dim():
    batch_size = 3
    params = jnp.linspace(0.0, 1.0, batch_size)
    envs = ParamEnv(offset=params)
    w = VmapEnvsWrapper(envs, batch_size=batch_size)
    bad_keys = jax.random.split(jax.random.key(0), batch_size + 1)
    with pytest.raises(ValueError):
        _ = w.init(bad_keys)


def test_pickle_serialization_of_state_and_info():
    batch_size = 3
    params = jnp.linspace(0.0, 1.0, batch_size)
    envs = ParamEnv(offset=params)
    w = VmapEnvsWrapper(envs, batch_size=batch_size)
    key = jax.random.key(0)
    state, info = w.init(key)
    blob = pickle.dumps((state, info))
    s2, i2 = pickle.loads(blob)
    assert jnp.allclose(s2, state)
    assert jnp.allclose(i2.obs, info.obs)


# -----------------------------------------------------------------------------
# Optional: Property-based sampling (offsets, batch sizes)
# -----------------------------------------------------------------------------


try:
    from hypothesis import given, settings
    from hypothesis import strategies as st
except Exception:  # pragma: no cover - optional dependency
    pytest.skip("hypothesis not installed", allow_module_level=True)


@given(
    batch_size=st.integers(min_value=1, max_value=5),
    seed=st.integers(min_value=0, max_value=2**31 - 1),
)
@settings(deadline=None, max_examples=20)
def test_prop_reset_step_shapes(batch_size, seed):
    params = jnp.linspace(-1.0, 1.0, batch_size)
    envs = ParamEnv(offset=params)
    w = VmapEnvsWrapper(envs, batch_size=batch_size)
    key = jax.random.key(seed)
    state, info = w.init(key)
    assert state is not None and info is not None
    assert info.obs.shape == (batch_size, 2)
    action = w.action_space.sample(key)
    s2, i2 = w.step(state, action)
    assert i2.obs.shape == (batch_size, 2)


def test_param_effect_applies_per_env_no_cross_mix():
    batch_size = 4
    offsets = jnp.array([0.0, 1.0, -1.0, 2.0], dtype=jnp.float32)
    envs = ParamEnv(offset=offsets)
    w = VmapEnvsWrapper(envs, batch_size=batch_size)
    key = jax.random.key(0)
    s, i = w.init(key)
    s2, i2 = w.step(s, jnp.zeros((batch_size, 2), dtype=jnp.float32))
    expected = i.obs + offsets.reshape((batch_size, 1))
    assert jnp.allclose(i2.obs, expected)


def test_unwrapped_returns_base_unwrapped():
    offsets = jnp.array([0.0, 1.0], dtype=jnp.float32)
    envs = ParamEnv(offset=offsets)
    w = VmapEnvsWrapper(envs, batch_size=2)
    assert w.unwrapped is envs.unwrapped



import jax
import jax.numpy as jnp

from envelope.wrappers.episode_statistics_wrapper import (
    EpisodeStatistics,
    EpisodeStatisticsWrapper,
)
from envelope.wrappers.truncation_wrapper import TruncationWrapper
from envelope.wrappers.vmap_wrapper import VmapWrapper
from tests.wrappers.helpers import StepCounterEnv


def test_init_creates_default_stats():
    env = StepCounterEnv()
    w = EpisodeStatisticsWrapper(env)
    key = jax.random.key(0)
    state, info = w.init(key)
    assert jnp.allclose(state.stats.reward, 0.0)
    assert jnp.allclose(state.stats.length, 0.0)
    assert hasattr(info, "stats")
    assert jnp.allclose(info.stats.reward, 0.0)
    assert jnp.allclose(info.stats.length, 0.0)


def test_step_accumulates_reward():
    env = StepCounterEnv()
    w = EpisodeStatisticsWrapper(env)
    key = jax.random.key(0)
    state, _ = w.init(key)
    actions = [0.1, 0.2, -0.1]
    for a in actions:
        state, info = w.step(state, jnp.asarray(a))
    assert jnp.allclose(state.stats.reward, sum(actions))
    assert jnp.allclose(info.stats.reward, sum(actions))


def test_step_increments_length():
    env = StepCounterEnv()
    w = EpisodeStatisticsWrapper(env)
    key = jax.random.key(0)
    state, _ = w.init(key)
    n_steps = 5
    for _ in range(n_steps):
        state, info = w.step(state, jnp.asarray(0.1))
    assert jnp.allclose(state.stats.length, n_steps)
    assert jnp.allclose(info.stats.length, n_steps)


def test_info_contains_stats_field():
    env = StepCounterEnv()
    w = EpisodeStatisticsWrapper(env)
    key = jax.random.key(0)
    state, info = w.init(key)
    assert hasattr(info, "stats")
    state, info = w.step(state, jnp.asarray(0.5))
    assert hasattr(info, "stats")
    assert isinstance(info.stats, EpisodeStatistics)


def test_reset_preserves_stats():
    env = StepCounterEnv()
    w = EpisodeStatisticsWrapper(env)
    key = jax.random.key(0)
    state, _ = w.init(key)
    for _ in range(3):
        state, _ = w.step(state, jnp.asarray(0.2))
    reward_before = state.stats.reward
    length_before = state.stats.length
    state, info = w.reset(state, key)
    assert jnp.allclose(state.stats.reward, reward_before)
    assert jnp.allclose(state.stats.length, length_before)
    assert jnp.allclose(info.stats.reward, reward_before)
    assert jnp.allclose(info.stats.length, length_before)
    assert w.observation_space.contains(info.obs)


def test_stats_persist_and_continue_after_reset():
    env = StepCounterEnv()
    w = EpisodeStatisticsWrapper(env)
    key = jax.random.key(0)
    state, _ = w.init(key)
    for _ in range(3):
        state, _ = w.step(state, jnp.asarray(0.1))
    state, _ = w.reset(state, key)
    for _ in range(2):
        state, _ = w.step(state, jnp.asarray(0.1))
    # Total length = 3 + 2 = 5, reward = 0.1*5 = 0.5
    assert jnp.allclose(state.stats.length, 5)
    assert jnp.allclose(state.stats.reward, 0.5)


def test_negative_rewards_accumulate_correctly():
    env = StepCounterEnv()
    w = EpisodeStatisticsWrapper(env)
    key = jax.random.key(0)
    state, _ = w.init(key)
    state, _ = w.step(state, jnp.asarray(1.0))
    state, _ = w.step(state, jnp.asarray(-0.5))
    assert jnp.allclose(state.stats.reward, 0.5)


def test_state_is_episode_statistics_state_with_inner_state():
    env = StepCounterEnv()
    w = EpisodeStatisticsWrapper(env)
    key = jax.random.key(0)
    state, _ = w.init(key)
    assert hasattr(state, "inner_state")
    assert hasattr(state, "stats")
    assert state.inner_state is not None


def test_state_unwrapped_reaches_inner_env_state():
    env = StepCounterEnv()
    w = EpisodeStatisticsWrapper(env)
    key = jax.random.key(0)
    state, _ = w.init(key)
    assert hasattr(state, "unwrapped")
    inner = state.unwrapped
    assert hasattr(inner, "env_state")


def test_observation_space_action_space_unchanged():
    env = StepCounterEnv()
    w = EpisodeStatisticsWrapper(env)
    assert w.observation_space is env.observation_space
    assert w.action_space is env.action_space


def test_observation_space_contains_after_init_and_step():
    env = StepCounterEnv()
    w = EpisodeStatisticsWrapper(env)
    key = jax.random.key(0)
    state, info = w.init(key)
    assert w.observation_space.contains(info.obs)
    state, info = w.step(state, jnp.asarray(0.1))
    assert w.observation_space.contains(info.obs)


def test_jit_init_step_loop():
    env = StepCounterEnv()
    w = EpisodeStatisticsWrapper(env)
    key = jax.random.key(0)
    n_steps = 4

    @jax.jit
    def run_n_steps(k):
        s, _ = w.init(k)

        def body(carry, _):
            next_s, _ = w.step(carry, jnp.asarray(0.1))
            return next_s, ()

        s, _ = jax.lax.scan(body, s, None, length=n_steps)
        return s.stats

    stats = run_n_steps(key)
    assert jnp.allclose(jnp.asarray(stats.length), n_steps)
    assert jnp.allclose(jnp.asarray(stats.reward), 0.4)


def test_composability_with_vmap_wrapper():
    batch_size = 3
    env = StepCounterEnv()
    w = EpisodeStatisticsWrapper(VmapWrapper(env, batch_size=batch_size))
    key = jax.random.key(0)
    state, _ = w.init(key)
    action = jnp.array([0.1, 0.2, 0.3])
    state, info = w.step(state, action)
    # Reward is batched; length may be scalar or batched depending on implementation
    assert jnp.asarray(state.stats.reward).shape == (batch_size,)
    assert jnp.allclose(jnp.asarray(state.stats.reward), action)


def test_composability_with_truncation_wrapper():
    env = StepCounterEnv(terminate_after=10)
    w = EpisodeStatisticsWrapper(TruncationWrapper(env, max_steps=5))
    key = jax.random.key(0)
    state, _ = w.init(key)
    for _ in range(6):
        state, info = w.step(state, jnp.asarray(0.1))
    # Truncation happens at step 5; stats still accumulate through truncation
    assert jnp.allclose(state.stats.length, 6)
    assert jnp.allclose(state.stats.reward, 0.6)



import jax
import jax.numpy as jnp
import pytest

from envelope.spaces import BatchedSpace
from envelope.wrappers.episode_statistics_wrapper import EpisodeStatisticsWrapper
from envelope.wrappers.pooled_init_vmap_wrapper import PooledInitVmapWrapper
from envelope.wrappers.truncation_wrapper import TruncationWrapper
from tests.wrappers.helpers import ScalarToyEnv, StepCounterEnv


def test_init_creates_batched_state():
    batch_size = 4
    env = ScalarToyEnv()
    w = PooledInitVmapWrapper(env, batch_size=batch_size, pool_size=4)
    key = jax.random.key(0)
    state, info = w.init(key)
    assert info.obs.shape == (batch_size,)


def test_init_last_final_is_nan_placeholder():
    batch_size = 2
    env = ScalarToyEnv()
    w = PooledInitVmapWrapper(env, batch_size=batch_size, pool_size=2)
    key = jax.random.key(0)
    state, info = w.init(key)
    # last_final should be info-shaped with NaN
    assert jnp.all(jnp.isnan(state.last_final.obs))
    assert jnp.all(jnp.isnan(jnp.asarray(state.last_final.reward).reshape(-1)))


def test_init_info_has_final_field():
    env = ScalarToyEnv()
    w = PooledInitVmapWrapper(env, batch_size=2, pool_size=2)
    key = jax.random.key(0)
    state, info = w.init(key)
    assert hasattr(info, "final")
    assert jnp.all(jnp.isnan(info.final.obs))


def test_step_non_done_envs_continue_normally():
    env = ScalarToyEnv()  # never done
    w = PooledInitVmapWrapper(env, batch_size=2, pool_size=2)
    key = jax.random.key(0)
    state, info = w.init(key)
    action = jnp.array([0.1, 0.2])
    state, info = w.step(state, action)
    # State should have progressed (obs = 0 + action)
    assert jnp.allclose(info.obs, jnp.array([0.1, 0.2]))


def test_step_done_envs_get_pool_states():
    env = StepCounterEnv(terminate_after=1)
    w = PooledInitVmapWrapper(env, batch_size=2, pool_size=4)
    key = jax.random.key(0)
    state, _ = w.init(key)
    action = jnp.array([0.1, 0.1])  # both take one step -> both done
    state, info = w.step(state, action)
    # Done envs get fresh init from pool; obs should be initial (0) from pool
    assert info.obs.shape == (2,)
    # Both envs terminated so both get pool inits (obs 0.0)
    assert jnp.allclose(info.obs, jnp.zeros(2))


def test_step_stores_terminal_info_in_last_final():
    env = StepCounterEnv(terminate_after=1)
    w = PooledInitVmapWrapper(env, batch_size=1, pool_size=1)
    key = jax.random.key(0)
    state, _ = w.init(key)
    action = jnp.array([0.5])
    state, info = w.step(state, action)
    # Terminal step: obs was 0+0.5=0.5, reward=0.5. last_final should have that terminal info
    assert jnp.allclose(state.last_final.obs, jnp.array([0.5]))
    assert jnp.allclose(state.last_final.reward, jnp.array(0.5))


def test_step_continue_envs_preserve_previous_last_final():
    env = StepCounterEnv(terminate_after=2)
    w = PooledInitVmapWrapper(env, batch_size=2, pool_size=2)
    key = jax.random.key(0)
    state, _ = w.init(key)
    # Step once: no env done yet (env_state 0.1 each)
    state, _ = w.step(state, jnp.array([0.1, 0.1]))
    # Step again: both hit terminate_after=2; terminal obs = 0.1+0.1 = 0.2 each
    state, info = w.step(state, jnp.array([0.1, 0.1]))
    # One more step: both get pool inits, neither done; last_final preserved
    state, _ = w.step(state, jnp.array([0.0, 0.0]))
    assert state.last_final.obs.shape == (2,)
    assert jnp.allclose(state.last_final.obs, jnp.array([0.2, 0.2]))
    assert jnp.allclose(state.last_final.reward, jnp.array([0.1, 0.1]))


def test_step_info_final_field():
    env = StepCounterEnv(terminate_after=1)
    w = PooledInitVmapWrapper(env, batch_size=1, pool_size=1)
    key = jax.random.key(0)
    state, info = w.init(key)
    assert hasattr(info, "final")
    state, info = w.step(state, jnp.array([0.5]))
    # For done env: final is terminal info
    assert hasattr(info, "final")
    assert jnp.allclose(info.final.obs, jnp.array([0.5]))


def test_reset_vmaps_inner_reset():
    batch_size = 3
    env = ScalarToyEnv()
    w = PooledInitVmapWrapper(env, batch_size=batch_size, pool_size=3)
    key = jax.random.key(0)
    state, info = w.init(key)
    state, info = w.reset(state, key)
    assert info.obs.shape == (batch_size,)
    assert w.observation_space.contains(info.obs)


def test_observation_space_is_batched_space():
    batch_size = 4
    env = ScalarToyEnv()
    w = PooledInitVmapWrapper(env, batch_size=batch_size, pool_size=4)
    assert isinstance(w.observation_space, BatchedSpace)
    assert w.observation_space.batch_size == batch_size
    assert w.observation_space.space == env.observation_space


def test_action_space_is_batched_space():
    batch_size = 4
    env = ScalarToyEnv()
    w = PooledInitVmapWrapper(env, batch_size=batch_size, pool_size=4)
    assert isinstance(w.action_space, BatchedSpace)
    assert w.action_space.batch_size == batch_size


def test_observation_space_contains_after_init_and_step():
    env = ScalarToyEnv()
    w = PooledInitVmapWrapper(env, batch_size=2, pool_size=2)
    key = jax.random.key(0)
    state, info = w.init(key)
    assert w.observation_space.contains(info.obs)
    state, info = w.step(state, jnp.array([0.1, -0.1]))
    assert w.observation_space.contains(info.obs)


def test_action_space_sample_contains():
    env = ScalarToyEnv()
    w = PooledInitVmapWrapper(env, batch_size=2, pool_size=2)
    key = jax.random.key(0)
    action = w.action_space.sample(key)
    assert w.action_space.contains(action)


@pytest.mark.parametrize("pool_size", [1, 2, 4])
def test_pool_size_parametrized(pool_size):
    env = ScalarToyEnv()
    w = PooledInitVmapWrapper(env, batch_size=2, pool_size=pool_size)
    key = jax.random.key(0)
    state, info = w.init(key)
    state, info = w.step(state, jnp.array([0.0, 0.0]))
    assert info.obs.shape == (2,)


def test_deterministic_given_same_key():
    env = ScalarToyEnv()
    w = PooledInitVmapWrapper(env, batch_size=2, pool_size=2)
    key = jax.random.key(42)
    state1, info1 = w.init(key)
    state2, info2 = w.init(key)
    assert jnp.allclose(info1.obs, info2.obs)
    state1, info1 = w.step(state1, jnp.array([0.1, 0.2]))
    state2, info2 = w.step(state2, jnp.array([0.1, 0.2]))
    assert jnp.allclose(info1.obs, info2.obs)


def test_different_keys_different_pool_states():
    """Different keys produce valid inits; shape matches. ScalarToyEnv init is deterministic (obs 0), so we only assert no crash and shape."""
    env = ScalarToyEnv()
    w = PooledInitVmapWrapper(env, batch_size=2, pool_size=2)
    s1, i1 = w.init(jax.random.key(0))
    s2, i2 = w.init(jax.random.key(1))
    assert s1.inner_state.shape == s2.inner_state.shape
    assert i1.obs.shape == i2.obs.shape == (2,)


def test_all_envs_done_simultaneously():
    env = StepCounterEnv(terminate_after=1)
    w = PooledInitVmapWrapper(env, batch_size=4, pool_size=4)
    key = jax.random.key(0)
    state, _ = w.init(key)
    action = jnp.array([0.1, 0.1, 0.1, 0.1])
    state, info = w.step(state, action)
    assert info.obs.shape == (4,)
    assert state.last_final.obs.shape == (4,)


def test_no_envs_done():
    env = ScalarToyEnv()
    w = PooledInitVmapWrapper(env, batch_size=3, pool_size=2)
    key = jax.random.key(0)
    state, _ = w.init(key)
    state, info = w.step(state, jnp.array([0.1, 0.2, 0.3]))
    assert jnp.allclose(info.obs, jnp.array([0.1, 0.2, 0.3]))
    assert jnp.all(jnp.isnan(state.last_final.obs))


def test_batch_size_one_pool_size_one():
    env = ScalarToyEnv()
    w = PooledInitVmapWrapper(env, batch_size=1, pool_size=1)
    key = jax.random.key(0)
    state, info = w.init(key)
    assert info.obs.shape == (1,)
    state, info = w.step(state, jnp.array([0.5]))
    assert info.obs.shape == (1,)


def test_pool_size_greater_than_batch_size():
    env = ScalarToyEnv()
    w = PooledInitVmapWrapper(env, batch_size=2, pool_size=8)
    key = jax.random.key(0)
    state, info = w.init(key)
    state, info = w.step(state, jnp.array([0.0, 0.0]))
    assert info.obs.shape == (2,)


def test_init_key_advances_each_step():
    env = ScalarToyEnv()
    w = PooledInitVmapWrapper(env, batch_size=1, pool_size=1)
    key = jax.random.key(0)
    state, _ = w.init(key)
    key_before = state.init_key
    state, _ = w.step(state, jnp.array([0.0]))
    key_after = state.init_key
    assert not jnp.allclose(key_before, key_after)


def test_jit_init_step():
    env = ScalarToyEnv()
    w = PooledInitVmapWrapper(env, batch_size=2, pool_size=2)
    key = jax.random.key(0)

    @jax.jit
    def run(k):
        s, i = w.init(k)
        a = w.action_space.sample(k)
        s, i = w.step(s, a)
        return i.obs.shape

    shape = run(key)
    assert shape == (2,)


def test_jax_lax_scan_multi_step_loop():
    env = ScalarToyEnv()
    w = PooledInitVmapWrapper(env, batch_size=2, pool_size=2)
    key = jax.random.key(0)
    state, _ = w.init(key)
    keys = jax.random.split(key, 5)
    actions = jnp.stack([w.action_space.sample(k) for k in keys])

    def body(carry, action):
        next_s, info = w.step(carry, action)
        return next_s, info.obs

    final_state, obs_stack = jax.lax.scan(body, state, actions)
    assert obs_stack.shape == (5, 2)


def test_composability_with_episode_statistics_wrapper():
    env = StepCounterEnv(terminate_after=2)
    w = EpisodeStatisticsWrapper(PooledInitVmapWrapper(env, batch_size=2, pool_size=2))
    key = jax.random.key(0)
    state, _ = w.init(key)
    for _ in range(4):
        state, _ = w.step(state, jnp.array([0.1, 0.1]))
    # Stats accumulate across pool resets; reward is batched
    assert jnp.asarray(state.stats.reward).shape == (2,)


def test_composability_with_truncation_wrapper():
    env = StepCounterEnv(truncate_after=2)
    w = PooledInitVmapWrapper(
        TruncationWrapper(env, max_steps=2),
        batch_size=2,
        pool_size=2,
    )
    key = jax.random.key(0)
    state, _ = w.init(key)
    state, info = w.step(state, jnp.array([0.1, 0.1]))
    state, info = w.step(state, jnp.array([0.1, 0.1]))
    # After truncation both envs get pool inits (ScalarToyEnv init obs = 0)
    assert info.obs.shape == (2,)
    assert jnp.allclose(info.obs, jnp.zeros(2))



"""Tests for envelope.wrappers.state_injection_wrapper.StateInjectionWrapper."""

import jax
import jax.numpy as jnp
import pytest

from envelope.wrappers.autoreset_wrapper import AutoResetWrapper
from envelope.wrappers.state_injection_wrapper import StateInjectionWrapper
from tests.wrappers.helpers import StepCounterEnv, StepState

# ============================================================================
# Tests: Core Functionality
# ============================================================================


class TestStateInjectionCoreFunctionality:
    """Test StateInjectionWrapper core functionality."""

    def test_init_delegates_to_inner_env(self):
        """Verify that init() calls inner env init."""
        env = StepCounterEnv()
        w = StateInjectionWrapper(env)
        key = jax.random.key(0)

        state, info = w.init(key)

        # Should get fresh state from inner env (StepCounterEnv starts at 0)
        assert state.inner_state is not None
        assert jnp.allclose(state.inner_state.env_state, jnp.array(0.0))
        assert state.inner_state.steps == 0
        assert jnp.allclose(info.obs, jnp.array(0.0))
        # reset_state is None until set_reset_state is called
        assert state.reset_state is None

    def test_set_reset_state_updates_state(self):
        """Verify that set_reset_state() updates the injected state."""
        env = StepCounterEnv()
        w = StateInjectionWrapper(env)
        key = jax.random.key(0)

        # Get initial state
        state, _ = w.init(key)

        # Set a custom reset state
        custom_state = StepState(env_state=jnp.array(42.0), steps=0)
        custom_obs = jnp.array(42.0)
        state = w.set_reset_state(state, custom_state, custom_obs)

        # Verify the state was updated
        assert jnp.allclose(state.reset_state.env_state, jnp.array(42.0))
        assert jnp.allclose(state.reset_obs, custom_obs)
        # inner_state should also be updated to match
        assert jnp.allclose(state.inner_state.env_state, jnp.array(42.0))

    def test_subsequent_reset_preserves_injected_state(self):
        """Verify that reset with existing state preserves the injected state."""
        env = StepCounterEnv()
        w = StateInjectionWrapper(env)
        key = jax.random.key(0)

        # Get initial state and set custom reset state
        state, _ = w.init(key)
        custom_state = StepState(env_state=jnp.array(42.0), steps=0)
        custom_obs = jnp.array(42.0)
        state = w.set_reset_state(state, custom_state, custom_obs)

        # Reset again, passing the current state (simulates auto-reset)
        key2 = jax.random.key(1)
        state2, info2 = w.reset(state, key2)

        # Should preserve the injected state
        assert jnp.allclose(state2.reset_state.env_state, jnp.array(42.0))
        assert jnp.allclose(info2.obs, custom_obs)

    def test_set_reset_state_overrides_existing(self):
        """Verify that set_reset_state() overrides existing injected state."""
        env = StepCounterEnv()
        w = StateInjectionWrapper(env)
        key = jax.random.key(0)

        # Get initial state and set first custom state
        state, _ = w.init(key)
        state = w.set_reset_state(
            state,
            StepState(env_state=jnp.array(1.0), steps=0),
            jnp.array(1.0),
        )

        # Set new reset state
        state = w.set_reset_state(
            state,
            StepState(env_state=jnp.array(99.0), steps=0),
            jnp.array(99.0),
        )

        # Should have new injected state
        assert jnp.allclose(state.reset_state.env_state, jnp.array(99.0))
        assert jnp.allclose(state.reset_obs, jnp.array(99.0))

    def test_step_updates_inner_state_but_preserves_reset_state(self):
        """Verify that step updates inner_state but keeps reset_state unchanged."""
        env = StepCounterEnv()
        w = StateInjectionWrapper(env)
        key = jax.random.key(0)

        # Get initial state and set custom reset state
        state, _ = w.init(key)
        custom_state = StepState(env_state=jnp.array(0.0), steps=0)
        custom_obs = jnp.array(0.0)
        state = w.set_reset_state(state, custom_state, custom_obs)

        # Take a step
        state, info = w.step(state, jnp.array(0.5))

        # inner_state should have progressed
        assert jnp.allclose(state.inner_state.env_state, jnp.array(0.5))
        assert state.inner_state.steps == 1
        # reset_state should be unchanged
        assert jnp.allclose(state.reset_state.env_state, jnp.array(0.0))
        assert state.reset_state.steps == 0

    def test_reset_with_state_but_no_reset_state_does_normal_reset(self):
        """Verify that reset with state but reset_state=None does normal reset."""
        env = StepCounterEnv()
        w = StateInjectionWrapper(env)
        key = jax.random.key(0)

        # Get initial state and take a step so inner_state differs from fresh reset
        state, _ = w.init(key)
        state, _ = w.step(state, jnp.array(5.0))
        assert jnp.allclose(state.inner_state.env_state, jnp.array(5.0))

        # Reset with this state (no reset_state set) - should delegate to inner env
        key2 = jax.random.key(1)
        state2, info2 = w.reset(state, key2)

        # Should have done a normal reset - inner_state is fresh from env
        assert jnp.allclose(state2.inner_state.env_state, jnp.array(0.0))
        assert state2.inner_state.steps == 0
        assert jnp.allclose(info2.obs, jnp.array(0.0))
        # reset_state stays None
        assert state2.reset_state is None

    def test_set_reset_state_raises_on_invalid_state(self):
        """Verify that set_reset_state raises when InjectedState not found."""
        env = StepCounterEnv()
        w = StateInjectionWrapper(env)

        # Create a state that doesn't contain InjectedState
        invalid_state = StepState(env_state=jnp.array(0.0), steps=0)

        with pytest.raises(ValueError, match="Could not find InjectedState"):
            w.set_reset_state(invalid_state, invalid_state, jnp.array(0.0))

    def test_reset_raises_on_partial_reset_state(self):
        """Verify that reset raises when only one of reset_state/reset_obs is set."""
        env = StepCounterEnv()
        w = StateInjectionWrapper(env)
        key = jax.random.key(0)

        # Create state with only reset_state set (not reset_obs)
        state_with_only_reset_state = w.InjectedState(
            inner_state=StepState(env_state=jnp.array(0.0), steps=0),
            reset_state=StepState(env_state=jnp.array(42.0), steps=0),
            reset_obs=None,
        )

        with pytest.raises(ValueError, match="must set both"):
            w.reset(state_with_only_reset_state, key)

        # Create state with only reset_obs set (not reset_state)
        state_with_only_reset_obs = w.InjectedState(
            inner_state=StepState(env_state=jnp.array(0.0), steps=0),
            reset_state=None,
            reset_obs=jnp.array(42.0),
        )

        with pytest.raises(ValueError, match="must set both"):
            w.reset(state_with_only_reset_obs, key)


# ============================================================================
# Tests: Composability with AutoResetWrapper
# ============================================================================


class TestStateInjectionWithAutoReset:
    """Test StateInjectionWrapper composability with AutoResetWrapper."""

    def test_auto_reset_returns_to_same_state(self):
        """Verify that auto-reset returns to the same injected state."""
        env = StepCounterEnv(terminate_after=2)
        inner_w = StateInjectionWrapper(env)
        w = AutoResetWrapper(inner_w)
        key = jax.random.key(0)

        # Reset and set a custom reset state
        state, _ = w.init(key)
        reset_state = StepState(env_state=jnp.array(42.0), steps=0)
        reset_obs = jnp.array(42.0)
        state = inner_w.set_reset_state(state, reset_state, reset_obs)

        # Step until termination triggers auto-reset
        state, _ = w.step(state, jnp.array(0.1))
        state, info = w.step(state, jnp.array(0.2))  # Terminates

        # After auto-reset, should be back to same injected state
        assert jnp.allclose(state.inner_state.reset_state.env_state, jnp.array(42.0))
        assert jnp.allclose(info.obs, reset_obs)

    def test_set_reset_state_with_autoreset_wrapper(self):
        """Verify that set_reset_state works through AutoResetWrapper."""
        env = StepCounterEnv(terminate_after=1)
        inner_w = StateInjectionWrapper(env)
        w = AutoResetWrapper(inner_w)
        key = jax.random.key(0)

        # Get initial state
        state, _ = w.init(key)

        # Set a custom reset state - just pass the outermost state
        custom_state = StepState(env_state=jnp.array(100.0), steps=0)
        custom_obs = jnp.array(100.0)
        state = inner_w.set_reset_state(state, custom_state, custom_obs)

        # Step to trigger termination → auto-reset
        state, info = w.step(state, jnp.array(0.1))

        # After auto-reset, should return to custom injected state
        assert jnp.allclose(state.inner_state.reset_state.env_state, jnp.array(100.0))
        assert jnp.allclose(info.obs, custom_obs)

    def test_multiple_auto_resets_preserve_state(self):
        """Verify that injected state persists through multiple auto-resets."""
        env = StepCounterEnv(terminate_after=1)
        inner_w = StateInjectionWrapper(env)
        w = AutoResetWrapper(inner_w)
        key = jax.random.key(0)

        # Get initial state and set custom reset state
        state, _ = w.init(key)
        custom_state = StepState(env_state=jnp.array(50.0), steps=0)
        custom_obs = jnp.array(50.0)
        state = inner_w.set_reset_state(state, custom_state, custom_obs)

        # Trigger multiple auto-resets
        for _ in range(5):
            state, info = w.step(state, jnp.array(0.1))
            # Each auto-reset should return to the same injected state
            assert jnp.allclose(
                state.inner_state.reset_state.env_state, jnp.array(50.0)
            )
            assert jnp.allclose(info.obs, custom_obs)

    def test_ued_style_outer_loop(self):
        """Test the UED-style outer loop pattern with set_reset_state."""
        env = StepCounterEnv(terminate_after=2)
        inner_w = StateInjectionWrapper(env)
        w = AutoResetWrapper(inner_w)
        key = jax.random.key(0)

        # Initial reset
        state, _ = w.init(key)

        # Simulate UED outer loop
        for outer_iter in range(3):
            # Sample a new "task" (starting state)
            task_value = jnp.array(float(outer_iter * 10))
            task_state = StepState(env_state=task_value, steps=0)
            task_obs = task_value

            # Set the reset state for this outer iteration
            # User just passes the outermost state
            state = inner_w.set_reset_state(state, task_state, task_obs)

            # Inner loop with auto-resets
            for inner_step in range(4):  # More than terminate_after to trigger resets
                state, info = w.step(state, jnp.array(0.1))
                # After auto-reset, obs should be task_obs
                # (on non-reset steps, obs will be different)

            # After inner loop, the reset_state should still be our task
            assert jnp.allclose(state.inner_state.reset_state.env_state, task_value)


# ============================================================================
# Tests: JIT Compatibility
# ============================================================================


class TestStateInjectionJITCompatibility:
    """Test StateInjectionWrapper JIT compatibility."""

    def test_jit_reset_and_step(self):
        """Verify that reset and step can be JIT compiled."""
        env = StepCounterEnv(terminate_after=2)
        w = AutoResetWrapper(StateInjectionWrapper(env))
        key = jax.random.key(0)

        @jax.jit
        def run_episode(k):
            s, _ = w.init(k)
            for _ in range(5):  # More steps than terminate_after to trigger resets
                s, info = w.step(s, jnp.array(0.1))
            return s, info

        state, info = run_episode(key)
        assert state is not None
        assert info is not None

    def test_jit_set_reset_state(self):
        """Verify JIT works with set_reset_state."""
        env = StepCounterEnv(terminate_after=1)
        inner_w = StateInjectionWrapper(env)
        w = AutoResetWrapper(inner_w)

        @jax.jit
        def run_with_state(k, reset_state, reset_obs):
            s, _ = w.init(k)
            # Set the reset state - just pass the outermost state
            s = inner_w.set_reset_state(s, reset_state, reset_obs)
            for _ in range(3):
                s, info = w.step(s, jnp.array(0.1))
            return s, info

        key = jax.random.key(0)
        custom_state = StepState(env_state=jnp.array(77.0), steps=0)
        custom_obs = jnp.array(77.0)

        state, info = run_with_state(key, custom_state, custom_obs)

        # Injected state should persist through auto-resets
        assert jnp.allclose(state.inner_state.reset_state.env_state, jnp.array(77.0))
        assert jnp.allclose(info.obs, custom_obs)

    def test_jit_ued_outer_loop(self):
        """Verify JIT works with the UED outer loop pattern."""
        env = StepCounterEnv(terminate_after=1)
        inner_w = StateInjectionWrapper(env)
        w = AutoResetWrapper(inner_w)

        @jax.jit
        def outer_iteration(state, task_value):
            task_state = StepState(env_state=task_value, steps=0)
            task_obs = task_value

            # Set reset state for this task - pass outermost state
            state = inner_w.set_reset_state(state, task_state, task_obs)

            # Run inner loop
            for _ in range(3):
                state, info = w.step(state, jnp.array(0.1))

            return state, info

        key = jax.random.key(0)
        state, _ = w.init(key)

        # Run multiple outer iterations
        for i in range(3):
            task_value = jnp.array(float(i * 10))
            state, info = outer_iteration(state, task_value)
            assert jnp.allclose(state.inner_state.reset_state.env_state, task_value)



from functools import cached_property
from typing import TYPE_CHECKING, Any, override

import jax
import jax.numpy as jnp
from craftax.craftax_env import make_craftax_env_from_name

from envelope import spaces as envelope_spaces
from envelope.adapters.gymnax_envelope import _convert_space as _convert_gymnax_space
from envelope.environment import Environment, Info, InfoContainer, State
from envelope.struct import Container, static_field
from envelope.typing import Key, PyTree, TypeAlias

if TYPE_CHECKING:
    from craftax.craftax.craftax_state import EnvParams as CraftaxEnvParamsOriginal
    from craftax.craftax.envs.craftax_pixels_env import (
        CraftaxPixelsEnv,
        CraftaxPixelsEnvNoAutoReset,
    )
    from craftax.craftax.envs.craftax_symbolic_env import (
        CraftaxSymbolicEnv,
        CraftaxSymbolicEnvNoAutoReset,
    )
    from craftax.craftax_classic.envs.craftax_pixels_env import (
        CraftaxClassicPixelsEnv,
        CraftaxClassicPixelsEnvNoAutoReset,
    )
    from craftax.craftax_classic.envs.craftax_state import (
        EnvParams as CraftaxClassicEnvParams,
    )
    from craftax.craftax_classic.envs.craftax_symbolic_env import (
        CraftaxClassicSymbolicEnv,
        CraftaxClassicSymbolicEnvNoAutoReset,
    )

    CraftaxEnvParams: TypeAlias = CraftaxEnvParamsOriginal | CraftaxClassicEnvParams
    CraftaxEnv: TypeAlias = (
        CraftaxPixelsEnv
        | CraftaxSymbolicEnv
        | CraftaxClassicPixelsEnv
        | CraftaxClassicSymbolicEnv
        | CraftaxPixelsEnvNoAutoReset
        | CraftaxSymbolicEnvNoAutoReset
        | CraftaxClassicPixelsEnvNoAutoReset
        | CraftaxClassicSymbolicEnvNoAutoReset
    )
else:
    CraftaxEnvParams: TypeAlias = Any
    CraftaxEnv: TypeAlias = Any


class CraftaxEnvelope(Environment):
    """
    Wrapper to convert a Craftax environment to a envelope environment.

    Craftax (mostly) uses the Gymnax interface, so the info object is only created on
    the first `step`. To keep structural equivalence between the `init` and `step`
    infos, we create a placeholder filled with `jnp.nan` that is returned on `init`.

    Args:
        craftax_env (CraftaxEnv): the Craftax environment, with baked-in
            `static_env_params`.
        env_params (CraftaxEnvParams): the environment parameters, which are passed to
            the Craftax environment's `reset` and `step` methods.
    """

    craftax_env: CraftaxEnv = static_field()
    env_params: PyTree = static_field()  # TODO: remove static marker as soon as craftax merges https://github.com/MichaelTMatthews/Craftax/pull/48

    @classmethod
    def from_name(
        cls,
        env_name: str,
        env_params: CraftaxEnvParams | None = None,
        env_kwargs: dict[str, Any] | None = None,
    ) -> "CraftaxEnvelope":
        """
        Create a `CraftaxEnvelope` from a name and keyword arguments.
        `env_kwargs` are passed to `craftax.craftax_env.make_craftax_env_from_name`.
        """
        env_kwargs = env_kwargs or {}
        if "max_timesteps" in env_kwargs:
            raise ValueError(
                "Cannot override 'max_timesteps' directly. "
                "Use TruncationWrapper for episode length control."
            )
        if "auto_reset" in env_kwargs:
            raise ValueError(
                "Cannot override 'auto_reset' directly. "
                "Use AutoResetWrapper for auto-reset behavior."
            )

        env_kwargs["auto_reset"] = False
        env = make_craftax_env_from_name(env_name, **env_kwargs)
        default_params = env.default_params.replace(max_timesteps=jnp.inf)

        env_params = env_params or default_params
        return cls(craftax_env=env, env_params=env_params)

    @property
    def default_max_steps(self) -> int:
        return int(self.craftax_env.default_params.max_timesteps)

    @cached_property
    def _craftax_info_placeholder(self) -> PyTree:
        # Note that the placeholder that is returned only has nan values where it's
        # dtype is a subdtype of float. TODO: Should we use empty_like?
        key = jax.random.key(0)
        _, state = self.craftax_env.reset(key, self.env_params)
        _, _, _, _, info = self.craftax_env.step(
            key,
            state,
            self.craftax_env.action_space(self.env_params).sample(key),
            self.env_params,
        )
        return jax.tree.map(lambda x: jnp.full_like(x, jnp.nan), info)

    @override
    def init(self, key: Key) -> tuple[State, Info]:
        key, subkey = jax.random.split(key)
        obs, env_state = self.craftax_env.reset(subkey, self.env_params)
        state = Container().update(key=key, env_state=env_state)
        info = InfoContainer(obs=obs, reward=0.0, terminated=False)
        info = info.update(info=self._craftax_info_placeholder)
        return state, info

    @override
    def step(self, state: State, action: PyTree) -> tuple[State, Info]:
        key, subkey = jax.random.split(state.key)
        obs, env_state, reward, done, env_info = self.craftax_env.step(
            subkey, state.env_state, action, self.env_params
        )
        state = state.update(key=key, env_state=env_state)
        info = InfoContainer(obs=obs, reward=reward, terminated=done)
        info = info.update(info=env_info)
        return state, info

    @override
    @cached_property
    def action_space(self) -> envelope_spaces.Space:
        return _convert_gymnax_space(self.craftax_env.action_space(self.env_params))

    @override
    @cached_property
    def observation_space(self) -> envelope_spaces.Space:
        return _convert_gymnax_space(
            self.craftax_env.observation_space(self.env_params)
        )



from functools import cached_property
from typing import override

import jax
import jax.numpy as jnp

from envelope.environment import Info, State
from envelope.spaces import BatchedSpace, Continuous, Discrete, Space, peel_batched
from envelope.typing import Key, PyTree
from envelope.wrappers.wrapper import Wrapper


def flatten_space(space: Space):
    def is_leaf(x):
        # Tuples containing only integers are shape tuples (leaves)
        # PyTreeSpace can only have tuples that contain at least a Space, so
        # tuples with only integers must be shape tuples from leaf spaces
        return isinstance(x, tuple) and all(isinstance(i, int) for i in x)

    shapes, treedef = jax.tree.flatten(space.shape, is_leaf=is_leaf)
    dims = [jnp.prod(jnp.asarray(shape)) for shape in shapes]
    return treedef, shapes, dims


def flatten_x(x: PyTree):
    leaves = jax.tree.leaves(x)
    xs = jax.tree.map(lambda x: jnp.asarray(x).reshape(-1), leaves)
    x = jnp.concatenate(xs, axis=0)
    return x


class FlattenObservationWrapper(Wrapper):
    @override
    def init(self, key: Key) -> tuple[State, Info]:
        state, info = self.env.init(key)
        info = info.update(obs=flatten_x(info.obs))
        return state, info

    @override
    def reset(self, state: State, key: Key) -> tuple[State, Info]:
        next_state, info = self.env.reset(state, key)
        return next_state, info.update(obs=flatten_x(info.obs))

    @override
    def step(self, state: State, action: PyTree) -> tuple[State, Info]:
        state, info = self.env.step(state, action)
        info = info.update(obs=flatten_x(info.obs))
        return state, info

    @override
    @cached_property
    def observation_space(self) -> Space:
        batch_dims, base = peel_batched(self.env.observation_space)

        def is_leaf(x):
            spaces = (Continuous, Discrete)
            return isinstance(x, spaces)

        spaces = jax.tree.leaves(base, is_leaf=is_leaf)
        obs_cls = type(spaces[0])

        if not all(isinstance(space, obs_cls) for space in spaces):
            raise ValueError("All spaces must be of the same type")

        if obs_cls == Continuous:
            lows = [jnp.asarray(s.low).reshape(-1) for s in spaces]
            highs = [jnp.asarray(s.high).reshape(-1) for s in spaces]
            low = jnp.concatenate(lows, axis=0)
            high = jnp.concatenate(highs, axis=0)
            space = Continuous(low=low, high=high)
        elif obs_cls == Discrete:
            ns = [jnp.asarray(s.n).reshape(-1) for s in spaces]
            n = jnp.concatenate(ns, axis=0)
            space = Discrete(n=n)
        else:
            raise ValueError(f"Unsupported space type: {obs_cls}")

        for batch_dim in batch_dims:
            space = BatchedSpace(space, batch_dim)
        return space
